from .pigpio_mocks import patch_pigpio_base, patch_pigpio_gpio, soft_frequencies
from vent.io.devices.pins import Pin, PWMOutput
from secrets import token_bytes

import pytest
import random


@pytest.mark.parametrize("seed", [token_bytes(8) for _ in range(2)])
@pytest.mark.parametrize("gpio", range(53))
def test_mode(patch_pigpio_gpio, seed, gpio):
    """__________________________________________________________________________________________________________TEST #1
     Tests the mode setting & getting methods of Pin
         - Initializes a Pin
         - Reads mode off pin (should be random)
         - Writes a mode to Pin
         - Reads the mode back from the Pin
         - Asserts that the first (random) mode read is a valid mode
         - Asserts that the second mode is the mode we set
    """
    random.seed(seed)
    mode = random.choice([key for key in Pin._PIGPIO_MODES.keys()])
    results = []
    pin = Pin(gpio)
    results.append(pin.mode)
    if mode == 'Fake Mode':
        with pytest.raises(ValueError):
            pin.mode = mode
    else:
        pin.mode = mode
        results.append(pin.mode)
        assert results[1] == mode
    assert results[0] in Pin._PIGPIO_MODES

    """__________________________________________________________________________________________________________
    """


def test_bad_mode_exception(patch_pigpio_gpio):
    """__________________________________________________________________________________________________________TEST #2
     Tests that an exception is thrown if an attempt is made to set an unrecognized mode
    """
    gpio = random.choice(range(53))
    mode = 'Fake Mode'
    pin = Pin(gpio)
    with pytest.raises(ValueError):
        pin.mode = mode
    """__________________________________________________________________________________________________________
    """


@pytest.mark.parametrize("gpio", range(31))
@pytest.mark.parametrize("level", [0, 1])
def test_read_write_toggle(patch_pigpio_gpio, gpio, level):
    """__________________________________________________________________________________________________________TEST #3
     Tests the toggle feature of a Pin
         - Initializes a random Pin (in User_GPIO)
         - Sets mode to 'OUTPUT'
         - Writes a random level to Pin
         - Reads the pin and stores in results
         - Writes the level to Pin again
         - Toggles the Pin
         - Reads the Pin and stores in results
         - Asserts that the result is [level, not level]
         - Checks that an exception is thrown if you try to write a bad value
    """
    results = []
    pin = Pin(gpio)
    pin.mode = 'OUTPUT'
    pin.write(level)
    results.append(pin.read())
    pin.write(level)
    pin.toggle()
    results.append(pin.read())
    assert results[0] == level
    assert results[1] is not level
    with pytest.raises(ValueError):
        pin.write(-1)
    """__________________________________________________________________________________________________________
    """


@pytest.mark.parametrize("seed", [token_bytes(8) for _ in range(128)])
def test_frequency(patch_pigpio_gpio, seed):
    """__________________________________________________________________________________________________________TEST #4
     Tests the frequency setter & getter properties, and checks that duty is not changed or something weird like that
         - Initializes a PWMOutput
         - Sets frequency
         - reads frequency and appends to results
         - reads duty and appends to results
         - set up a condition that should never happen and check that the driver recovers (lie: hardware_enabled = True)
    """
    random.seed(seed)
    gpio = random.choice(range(31))
    offspec = False
    if gpio not in PWMOutput._HARDWARE_PWM_PINS:
        if random.getrandbits(3) == 7:
            frequency = random.choice(soft_frequencies)
        else:
            frequency = random.randint(1, 10000)
            while frequency in soft_frequencies:
                frequency = random.randint(1, 10000)
            offspec = True
    else:
        frequency = random.randint(1, 20000000)
    results = []
    pin = PWMOutput(gpio)
    if offspec:
        with pytest.raises(RuntimeWarning):
            pin.frequency = frequency
    else:
        pin.frequency = frequency
        results.append(pin.frequency)
        results.append(pin.read())
        assert results[0] == frequency
        assert results[1] == 0
    if not pin.hardware_enabled:
        with pytest.raises(Exception):
            pin._hardware_enabled = True
            pin.frequency = 20000
    """__________________________________________________________________________________________________________
    """


@pytest.mark.parametrize("gpio", random.sample(range(31), 16))
@pytest.mark.parametrize("duty", random.sample([x/100 for x in range(101)], k=32))
def test_duty(patch_pigpio_gpio, gpio, duty):
    """__________________________________________________________________________________________________________TEST #5
     Tests the duty setter & getter properties (and synonym write())
         - Initializes a PWMOutput
         - Reads frequency and appends to results
         - Writes duty
         - Reads duty and appends to results
         - Reads freqency and appends to results
         - Asserts frequency was not changed
         - Asserts duty read from pin matches input
    """
    results = []
    pin = PWMOutput(gpio)
    results.append(pin.frequency)
    pin.write(duty)
    results.append(pin.duty)
    results.append(pin.frequency)
    assert results[0] == results[2]
    assert round(results[1], 2) == duty
    with pytest.raises(ValueError):
        pin.duty = -1
    """__________________________________________________________________________________________________________
    """





""" Decorators for dangerous functions

"""
from vent.common.logging import log_exception
import traceback
import functools


MAX_STACK_DEPTH = 20


def locked(func):
    """
    Wrapper to use as decorator, handle lock logic for a
    @property

    Arguments:
        func (callable): function to wrap
    """

    # define a function to return that wraps the inner
    # function around some standardized error handling
    # logic

    # use .wraps to preserve inner method identity
    @functools.wraps(func)
    def lock_wrapper(self, *args, **kwargs):
        ret = None

        try:
            self.lock.acquire()
            ret = func(self, *args, **kwargs)
        except Exception as e:
            log_exception(
                e,
                traceback.TracebackException.from_exception(e, limit=MAX_STACK_DEPTH)
            )
        finally:
            self.lock.release()

        return ret

    return lock_wrapper


def pigpio_command(func):
    @functools.wraps(func)
    def exception_catcher(self, *args, **kwargs):
        result = None
        try:
            result = func(self, *args, **kwargs)
        except Exception as e:
            log_exception(e, traceback.TracebackException.from_exception(e))
        return result

    return exception_catcher



import multiprocessing
import time

from vent.coordinator import rpc


class ProcessManager:
    # Functions:
    def __init__(self, sim_mode, startCommandLine=None, maxHeartbeatInterval=None):
        self.sim_mode = sim_mode
        self.command_line = None  # TODO: what is this?
        self.max_heartbeat_interval = None
        self.previous_timestamp = None
        self.child_process = None
        self.serve_event = multiprocessing.Event()
        self.serve_event.clear()
        self.timeout = 5
        # TODO: if child process exists, need to reconnect it
        self.start_process()
        #time.sleep(1)

    def __del__(self):
        self.try_stop_process()

    def start_process(self):
        if self.child_process is not None:
            # Child process already started
            return
        self.serve_event.clear()
        self.child_process = multiprocessing.Process(target=rpc.rpc_server_main,
                                                     kwargs=
                                                     {
                                                         'sim_mode':self.sim_mode,
                                                         'serve_event':self.serve_event
                                                     })
        # self.child_process.daemon = True
        self.child_process.start()
        self.child_pid = self.child_process.pid
        self.serve_event.wait(self.timeout)

    def try_stop_process(self):
        if self.child_process is not None:
            # print(f'kill process {self.child_pid}')
            self.child_process.kill()
            while self.child_process.is_alive():
                time.sleep(0.01)
            self.child_process = None
            self.child_pid = None

    def restart_process(self):
        self.try_stop_process()
        self.start_process()

    def heartbeat(self, timestamp):
        # TODO: if no heartbeat in maxInterval restart
        pass

    def __del__(self):
        try:
            self.stop_process()
        except AttributeError:
            pass




import numpy as np
from PySide2 import QtWidgets, QtCore

from vent.gui import styles, mono_font
from vent.gui.widgets.components import EditableLabel, DoubleSlider


class Control(QtWidgets.QWidget):

    value_changed = QtCore.Signal(float)

    def __init__(self, value):
        super(Control, self).__init__()

        self.name = value.name
        self.units = value.units
        self.abs_range = value.abs_range
        self.safe_range = value.safe_range
        self.value = value.default
        self.decimals = value.decimals

        self.init_ui()


    def init_ui(self):
        self.layout = QtWidgets.QGridLayout()
        # self.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
        #                           QtWidgets.QSizePolicy.Expanding)

        # Value, Controller
        #        min,   max
        # Name
        # Units

        self.value_label = EditableLabel()
        self.value_label.setStyleSheet(styles.CONTROL_VALUE)
        self.value_label.label.setFont(mono_font())
        self.value_label.lineEdit.setFont(mono_font())
        self.value_label.label.setAlignment(QtCore.Qt.AlignRight)
        #self.value_label.setMargin(0)
        self.value_label.setContentsMargins(0,0,0,0)

        # # set max size based on range
        # # FIXME: TEMPORARY HACK - get sizing to work intelligibly with the dial
        # #n_ints = len(str(self.abs_range[1]))
        # n_ints = 3
        if self.decimals > 0:
            # don't just add decimals b/c we add another for the decimal itself
            # but don't want to +1 if decimals = 0!
            n_ints = int(len(str(self.abs_range[1]))+self.decimals+1)
        else:
            n_ints = int(len(str(self.abs_range[1])))

        # to simplify things for now, if we're supposed to display _more_
        # than the 5 we assume (eg. 100.0), then use the n_ints
        # otherwise just use 5
        if n_ints <= 5:
            n_ints = 5
        self.value_label.setFixedWidth(n_ints*styles.VALUE_SIZE*.6)
        # self.value_label.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
        #                             QtWidgets.QSizePolicy.Maximum)



        self.name_label = QtWidgets.QLabel()
        self.name_label.setStyleSheet(styles.CONTROL_NAME)
        self.name_label.setText(self.name)
        self.name_label.setWordWrap(True)
        self.name_label.setAlignment(QtCore.Qt.AlignRight)
        # self.name_label.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
        #                             QtWidgets.QSizePolicy.Expanding)


        self.units_label = QtWidgets.QLabel()
        self.units_label.setStyleSheet(styles.CONTROL_UNITS)
        self.units_label.setText(self.units)
        self.units_label.setAlignment(QtCore.Qt.AlignRight)
        self.units_label.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                    QtWidgets.QSizePolicy.Expanding)

        # Expand drawer button
        self.toggle_button = QtWidgets.QToolButton(checkable=True,
                                                   checked=False)
        self.toggle_button.setStyleSheet(styles.DISPLAY_NAME)

        self.toggle_button.toggled.connect(self.toggle_control)
        self.toggle_button.setToolButtonStyle(
            QtCore.Qt.ToolButtonIconOnly
        )
        self.toggle_button.setSizePolicy(QtWidgets.QSizePolicy.Maximum,
                                         QtWidgets.QSizePolicy.Expanding)
        self.toggle_button.setArrowType(QtCore.Qt.LeftArrow)

        ###
        # layout
        self.layout.addWidget(self.value_label, 0, 0, 2, 1, alignment=QtCore.Qt.AlignVCenter | QtCore.Qt.AlignRight)
        #self.layout.addWidget(self.dial, 0, 1, 2, 2, alignment=QtCore.Qt.AlignVCenter)
        #self.layout.addWidget(self.slider_min, 2, 1, 1, 1)
        #self.layout.addWidget(self.slider_max, 2, 2, 1, 1)
        self.layout.addWidget(self.name_label, 0, 1, 1, 1)
        self.layout.addWidget(self.units_label, 1, 1, 1, 1)
        self.layout.addWidget(self.toggle_button, 0, 2, 2, 1, alignment=QtCore.Qt.AlignRight)

        self.setLayout(self.layout)


        ################
        # Create initially hidden widgets

        # Min value - slider - max value

        self.slider_layout = QtWidgets.QHBoxLayout()

        self.slider_min = QtWidgets.QLabel()
        self.slider_min.setText(str(np.round(self.abs_range[0], self.decimals)))
        self.slider_min.setAlignment(QtCore.Qt.AlignLeft)
        self.slider_min.setFont(mono_font())
        self.slider_min.setStyleSheet(styles.CONTROL_LABEL)
        self.slider_min.setMargin(0)
        self.slider_min.setSizePolicy(QtWidgets.QSizePolicy.Maximum,
                                    QtWidgets.QSizePolicy.Maximum)


        self.slider_max = QtWidgets.QLabel()
        self.slider_max.setText(str(np.round(self.abs_range[1], self.decimals)))
        self.slider_max.setAlignment(QtCore.Qt.AlignRight)
        self.slider_max.setFont(mono_font())
        self.slider_max.setStyleSheet(styles.CONTROL_LABEL)
        self.slider_max.setMargin(0)
        self.slider_max.setSizePolicy(QtWidgets.QSizePolicy.Maximum,
                                    QtWidgets.QSizePolicy.Maximum)

        self.slider = DoubleSlider(decimals=self.decimals, orientation=QtCore.Qt.Orientation.Horizontal)
        self.slider.setMinimum(self.abs_range[0])
        self.slider.setMaximum(self.abs_range[1])
        #self.slider.setContentsMargins(0,0,0,0)
        self.slider.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                  QtWidgets.QSizePolicy.Maximum)
        
        self.slider_layout.addWidget(self.slider_min)
        self.slider_layout.addWidget(self.slider)
        self.slider_layout.addWidget(self.slider_max)

        self.slider_frame = QtWidgets.QFrame()
        self.slider_frame.setLayout(self.slider_layout)
        self.slider_frame.setVisible(False)

        ###
        # set signals
        self.value_label.textChanged.connect(self.update_value)
        self.slider.doubleValueChanged.connect(self.update_value)


        self.update_value(self.value)

    @QtCore.Slot(bool)
    def toggle_control(self, state):
        if state == True:
            self.toggle_button.setArrowType(QtCore.Qt.DownArrow)
            self.layout.addWidget(self.slider_frame, 3, 0, 1, 3)
            self.slider_frame.setVisible(True)
            #self.adjustSize()
        else:
            self.toggle_button.setArrowType(QtCore.Qt.LeftArrow)
            self.layout.removeWidget(self.slider_frame)
            self.slider_frame.setVisible(False)
            #self.adjustSize()


    def update_value(self, new_value):
        if isinstance(new_value, str):
            new_value = float(new_value)

        if (new_value <= self.abs_range[1]) and (new_value >= self.abs_range[0]):
            self.value = new_value

            self.value_changed.emit(self.value)
        else:
            # TODO: Log this
            pass

        # still draw regardless in case an invalid value was given
        value_str = str(np.round(self.value, self.decimals))
        self.value_label.setText(value_str)

        self.slider.setValue(self.value)


from PySide2 import QtGui

SUBWAY_COLORS = {
    'blue': "#0039A6",
    'lime': "#6CBE45",
    'gray': "A7A9AC",
    'orange': "#FF6319",
    'yellow': "#FCCC0A",
    'red': "#EE352E",
    'green': "#00933C",
    'purple': "#B933AD"
}

BACKGROUND_COLOR = "#111111"
BOX_BACKGROUND = "#333333"
TEXT_COLOR = "#EEEEEE"
CONTROL_BACKGROUND = "#EEEEEE"
CONTROL_TEXT = BACKGROUND_COLOR
HANDLE_HEIGHT = 10
SLIDER_WIDTH = 80
SLIDER_HEIGHT = 50
INDICATOR_WIDTH = SLIDER_WIDTH/3
SLIDER_COLOR = TEXT_COLOR
INDICATOR_COLOR = SUBWAY_COLORS['blue']
ALARM_COLOR = "#FF0000"

DIVIDER_COLOR = "#FFFFFF"
DIVIDER_COLOR_DARK = BOX_BACKGROUND

VALUE_SIZE = 72
NAME_SIZE = 36
UNIT_SIZE = 18
TICK_SIZE = 12

MONITOR_UPDATE_INTERVAL = 0.5
"""
(float): inter-update interval (seconds) for :class:`~vent.gui.widgets.monitor.Monitor`
"""

GLOBAL = f"""
QWidget {{
    font-size: 20pt;
    color: {TEXT_COLOR};
}}
"""

RANGE_SLIDER = f"""
QSlider {{
    font-size : {TICK_SIZE}px;
}}

QSlider::groove:vertical {{
    border: 1px solid #FFFFFF;
    width: {SLIDER_WIDTH}px;
}}

QSlider::handle:vertical {{
    height: {HANDLE_HEIGHT}px;
    width: 20px;
    margin: 0px -20px 0px px;
    background-color: {SLIDER_COLOR};
}}

QSlider::groove:horizontal {{
    margin: 0px, 0px, 30px, 0px;
}}

QSlider::handle:horizontal {{

}}

"""


# RANGE_SLIDER = """
# QSlider {{
#     font-size: 12px;
# }}
# QSlider::groove:vertical {{
# border: 1px solid #FFFFFF;
# width: {slider_width}px;
# }}
# QSlider::handle:vertical {{
# height: {height}px;
# width: 20px;
# margin: 0px -20px 0px px;
# background-color: {slider_color};
# }}
#
# """.format(slider_width=INDICATOR_WIDTH,
#     height=HANDLE_HEIGHT,
#            slider_color=SLIDER_COLOR)



DISPLAY_VALUE =  """
QLabel {{ 
    color: {textcolor}; 
    font-size: {value_size}pt;
}}""".format(textcolor=TEXT_COLOR,
             value_size=VALUE_SIZE)

DISPLAY_VALUE_ALARM =  """
QLabel { 
    color: #ff0000; 
    font-size: 72pt; 
    font-weight: bold;
}"""

DISPLAY_NAME = """
QLabel {{ 
    color: {textcolor}; 
    font-size: {name_size}pt;
}}""".format(textcolor=TEXT_COLOR,
             name_size=NAME_SIZE)

DISPLAY_NAME_ALARM = """
QLabel {{ 
    color: #ff0000; 
    font-size: {name_size}pt;
}}""".format(name_size=NAME_SIZE)

DISPLAY_UNITS = """
QLabel {{ 
    color: {textcolor}; 
    font-size: {unit_size}pt;
}}""".format(textcolor=TEXT_COLOR,
             unit_size=UNIT_SIZE)

DISPLAY_UNITS_ALARM = f"""
QLabel {{ 
    color: #ff0000; 
    font-size: {UNIT_SIZE}pt;
}}"""

DISPLAY_WIDGET = """
border-bottom: 2px solid white;
"""

CONTROL_LABEL = f"""
QLabel {{
    font-size: 12px;
    color: {BACKGROUND_COLOR};
}}
"""

CONTROL_VALUE =  f"""
QLabel {{ 
    color: {BACKGROUND_COLOR}; 
    font-size: {VALUE_SIZE}pt;
}}
QLineEdit {{ 
    color: {TEXT_COLOR}; 
    font-size: {VALUE_SIZE}pt;
}}

"""

CONTROL_BOX = f"""
QGroupBox {{
    background-color: {TEXT_COLOR};
    border: 1px solid #000000;
    border-radius: 4px;
    margin-top: 20px;
}}

QGroupBox::title {{
  subcontrol-origin: margin;
  subcontrol-position: top left;
  left: 3px;
  top: -5px;
}}
"""

CONTROL_NAME = f"""
QLabel {{ 
    color: {BACKGROUND_COLOR}; 
    font-size: {NAME_SIZE}pt;
}}"""

CONTROL_UNITS = f"""
QLabel {{ 
    color: {BACKGROUND_COLOR}; 
    font-size: {UNIT_SIZE}pt;
}}"""

TITLE_STYLE = """
font-size: 32pt;
color: {text_color};
text-align: left;
""".format(text_color=TEXT_COLOR)



STATUS_NORMAL = f"""
QFrame {{
    background-color: {BACKGROUND_COLOR};
    color: {TEXT_COLOR};
}}
"""

STATUS_WARN = f"""
QFrame {{
    background-color: {SUBWAY_COLORS['orange']};
    color: {TEXT_COLOR};
}}
"""

STATUS_ALARM = f"""
QFrame {{
    background-color: {SUBWAY_COLORS['red']};
    color: {TEXT_COLOR};
}}
"""

HEARTBEAT_NORMAL = f"""
QRadioButton::indicator {{
    background: qradialgradient(cx:0, cy:0, radius:1, fx:0.5, fy:0.5, stop:0 white, stop:1 {SUBWAY_COLORS['blue']});
    border-radius: 5px;
}}

QLabel {{
    color: {TEXT_COLOR};
}}
"""

HEARTBEAT_ALARM = f"""
QRadioButton::indicator {{
    background: qradialgradient(cx:0, cy:0, radius:1, fx:0.5, fy:0.5, stop:0 white, stop:1 {SUBWAY_COLORS['red']});
    border-radius: 5px;
}}

QLabel {{
    color: {SUBWAY_COLORS['red']};
}}
"""

STATUS_BOX = f"""
QWidget {{
    background-color: {BACKGROUND_COLOR};
}}
"""

DARK_THEME =  f"""
/*
adapted from https://github.com/gmarull/qtmodern/blob/master/qtmodern/resources/style.qss
*/

/*
 * QGroupBox
 */


QGroupBox {{
  background-color: palette(alternate-base);
  border: 1px solid palette(midlight);
  margin-top: 25px;
}}

QGroupBox::title {{
    background-color: transparent;
}}

QGroupBox#CONTROLBOX {{
    background-color: {CONTROL_BACKGROUND};
}}

/*
 * QToolBar
 */

QToolBar {{
  border: none;
}}

/*
 * QTabBar
 */

QTabBar{{
  background-color: transparent;
}}

QTabBar::tab{{
  padding: 4px 6px;
  background-color: transparent;
  border-bottom: 2px solid transparent;
}}

QTabBar::tab:selected, QTabBar::tab:hover {{
  color: palette(text);
  border-bottom: 2px solid palette(highlight);
}}

QTabBar::tab:selected:disabled {{
  border-bottom: 2px solid palette(light);
}}

/*
 * QScrollBar
 */

QScrollBar:vertical {{
  background: palette(base);
  border-top-right-radius: 2px;
  border-bottom-right-radius: 2px;
  width: 16px;
  margin: 0px;
}}

QScrollBar::handle:vertical {{
  background-color: palette(alternate-base);
  border-radius: 2px;
  min-height: 20px;
  margin: 2px 4px 2px 4px;
}}

QScrollBar::handle:vertical:hover, QScrollBar::handle:horizontal:hover, QScrollBar::handle:vertical:pressed, QScrollBar::handle:horizontal:pressed {{
  background-color:palette(midlight);
}}

QScrollBar::add-line:vertical {{
  background: none;
  height: 0px;
  subcontrol-position: right;
  subcontrol-origin: margin;
}}

QScrollBar::sub-line:vertical {{
  background: none;
  height: 0px;
  subcontrol-position: left;
  subcontrol-origin: margin;
}}

QScrollBar:horizontal{{
  background: palette(base);
  height: 16px;
  margin: 0px;
}}

QScrollBar::handle:horizontal {{
  background-color: palette(alternate-base);
  border-radius: 2px;
  min-width: 20px;
  margin: 4px 2px 4px 2px;
}}


QScrollBar::add-line:horizontal {{
  background: none;
  width: 0px;
  subcontrol-position: bottom;
  subcontrol-origin: margin;
}}

QScrollBar::sub-line:horizontal {{
  background: none;
  width: 0px;
  subcontrol-position: top;
  subcontrol-origin: margin;
}}

/*
 * QScrollArea
 */

QScrollArea {{
  border-style: none;
}}

QScrollArea > QWidget > QWidget {{
  background-color: palette(alternate-base);
}}

/*
 * QSlider
 */

QSlider::handle:horizontal {{
  border-radius: 5px;
  background-color: palette(light);
  max-height: 20px;
}}

QSlider::add-page:horizontal {{
  background: palette(base);
}}

QSlider::sub-page:horizontal {{
  background: palette(highlight);
}}

QSlider::sub-page:horizontal:disabled {{
  background-color: palette(light);
}}

QTableView {{
  background-color: palette(link-visited);
  alternate-background-color: palette(midlight);
}}

QWidget {{
    font-size: 20pt;
}}"""


def set_dark_palette(app):
    """ Apply Dark Theme to the Qt application instance.

    borrowed from https://github.com/gmarull/qtmodern/blob/master/qtmodern/styles.py
        Args:
            app (QApplication): QApplication instance.
    """

    darkPalette = QtGui.QPalette()

    # base
    darkPalette.setColor(QtGui.QPalette.WindowText, QtGui.QColor(TEXT_COLOR))
    darkPalette.setColor(QtGui.QPalette.Button, QtGui.QColor(53, 53, 53))
    darkPalette.setColor(QtGui.QPalette.Light, QtGui.QColor(180, 180, 180))
    darkPalette.setColor(QtGui.QPalette.Midlight, QtGui.QColor(90, 90, 90))
    darkPalette.setColor(QtGui.QPalette.Dark, QtGui.QColor(35, 35, 35))
    darkPalette.setColor(QtGui.QPalette.Text, QtGui.QColor(180, 180, 180))
    darkPalette.setColor(QtGui.QPalette.BrightText, QtGui.QColor(180, 180, 180))
    darkPalette.setColor(QtGui.QPalette.ButtonText, QtGui.QColor(180, 180, 180))
    darkPalette.setColor(QtGui.QPalette.Base, QtGui.QColor(42, 42, 42))
    darkPalette.setColor(QtGui.QPalette.Window, QtGui.QColor(BACKGROUND_COLOR))
    darkPalette.setColor(QtGui.QPalette.Shadow, QtGui.QColor(20, 20, 20))
    darkPalette.setColor(QtGui.QPalette.Highlight, QtGui.QColor(42, 130, 218))
    darkPalette.setColor(QtGui.QPalette.HighlightedText, QtGui.QColor(180, 180, 180))
    darkPalette.setColor(QtGui.QPalette.Link, QtGui.QColor(56, 252, 196))
    darkPalette.setColor(QtGui.QPalette.AlternateBase, QtGui.QColor(BOX_BACKGROUND))
    darkPalette.setColor(QtGui.QPalette.ToolTipBase, QtGui.QColor(53, 53, 53))
    darkPalette.setColor(QtGui.QPalette.ToolTipText, QtGui.QColor(180, 180, 180))
    darkPalette.setColor(QtGui.QPalette.LinkVisited, QtGui.QColor(80, 80, 80))

    # disabled
    darkPalette.setColor(QtGui.QPalette.Disabled, QtGui.QPalette.WindowText,
                         QtGui.QColor(127, 127, 127))
    darkPalette.setColor(QtGui.QPalette.Disabled, QtGui.QPalette.Text,
                         QtGui.QColor(127, 127, 127))
    darkPalette.setColor(QtGui.QPalette.Disabled, QtGui.QPalette.ButtonText,
                         QtGui.QColor(127, 127, 127))
    darkPalette.setColor(QtGui.QPalette.Disabled, QtGui.QPalette.Highlight,
                         QtGui.QColor(80, 80, 80))
    darkPalette.setColor(QtGui.QPalette.Disabled, QtGui.QPalette.HighlightedText,
                         QtGui.QColor(127, 127, 127))

    app.setPalette(darkPalette)

    return app



#########################
# Imports

# python standard libraries
import os
from collections import OrderedDict as odict
from PySide2 import QtGui
from vent.common.values import ValueName, SENSOR
from vent.gui import styles

LIMIT_GUI_INSTANCE = True
"""
(bool): whether there hsould only be one GUI instance at a time. disabled during testing.
"""

def limit_gui(limit=None):
    if limit is None:
        return globals()['LIMIT_GUI_INSTANCE']
    else:
        globals()['LIMIT_GUI_INSTANCE'] = limit


PLOTS = odict({
    ValueName.PRESSURE: SENSOR[ValueName.PRESSURE].to_dict(),
    ValueName.TEMP: SENSOR[ValueName.TEMP].to_dict(),
    ValueName.HUMIDITY: SENSOR[ValueName.HUMIDITY].to_dict()
})
"""
Values to plot.

Should have the same key as some key in :data:`~.defaults.SENSOR`. If it does,
it will be mutually connected to the resulting :class:`.gui.widgets.Monitor_Value`
such that the set limit range is updated when the horizontal bars on the plot are updated.::

    {
        'name' (str): title of plot,
        'abs_range' (tuple): absolute limit of plot range,
        'safe_range' (tuple): safe range, will be discolored outside of this range,
        'color' (str): hex color of line (like "#FF0000")
    }
"""

PLOTS[ValueName.PRESSURE]['color'] = styles.SUBWAY_COLORS['orange']
PLOTS[ValueName.TEMP]['color'] = styles.SUBWAY_COLORS['red']
PLOTS[ValueName.HUMIDITY]['color'] = styles.SUBWAY_COLORS['blue']


########################

_GUI_INSTANCE = None

def set_gui_instance(instance):
    """
    Store the current instance of the GUI

    Arguments:
        instance (:class:`~.vent.gui.main.Vent_Gui`)
    """
    globals()['_GUI_INSTANCE'] = instance


def get_gui_instance():
    """
    Retreive the currently running instance of the GUI

    Returns:
        :class:`~.vent.gui.main.Vent_Gui`
    """
    return globals()['_GUI_INSTANCE']

###########
# Load a monospace font for displaying numbers
# Want to load an explicit font because computing the hint to find the default mono font is expensive

_MONO_FONT = None
def mono_font():
    """
    module function to return a :class:`PySide2.QtGui.QFont` to use as the mono font.

    use this instead of just making because :class:`PySide2.QtGui.QFontDatabase` can't be instantiated before the
    :class:`PySide2.QtWidgets.QApplication` is instantiated, so we load the font after the app
    """
    if globals()['_MONO_FONT'] is None:
        load_mono_font()

    return globals()['_MONO_FONT']

def load_mono_font():
    """
    Load the monospaced font and set the module-global :data:`_MONO_FONT` object.

    .. note::

        Must be called after :class:`PySide2.QtWidgets.QApplication` is instantiated!

    """
    try:
        # first try to load fira code for monospace font
        external_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'external')
        font_db = QtGui.QFontDatabase()
        font_db.addApplicationFont(os.path.join(external_dir, 'FiraCode-Regular.otf'))
        font_db.addApplicationFont(os.path.join(external_dir, 'FiraCode-Bold.otf'))
        mono_font = QtGui.QFont('Fira Code')
    except:   # pragma: no cover
        # if that fails, try to load liberation mono
        # TODO: Log this
        try:
            mono_font = QtGui.QFont('Liberation Mono')

        except:
            # otherwise get the system default mono font
            mono_font = QtGui.QFont()
            mono_font.setStyleHint(QtGui.QFont.Monospace)

    globals()['_MONO_FONT'] = mono_font


from vent.gui.main import Vent_Gui, launch_gui, get_gui_instance



import logging
import pickle
import socket
import xmlrpc.client
from xmlrpc.server import SimpleXMLRPCServer

import vent.controller.control_module
from vent.common.logging import init_logger

default_addr = 'localhost'
default_port = 9533
default_timeout = 10
socket.setdefaulttimeout(default_timeout)

remote_controller = None


def get_sensors():
    # left as example of how to get loggers within these callbacks
    #logger = logging.getLogger(__name__)
    #logger.info('remote runnnnnnn')
    res = remote_controller.get_sensors()
    return pickle.dumps(res)


# def get_active_alarms():
#     res = remote_controller.get_active_alarms()
#     return pickle.dumps(res)
#
#
# def get_logged_alarms():
#     res = remote_controller.get_logged_alarms()
#     return pickle.dumps(res)


def set_control(control_setting):
    args = pickle.loads(control_setting.data)
    remote_controller.set_control(args)


def get_control(control_setting_name):
    args = pickle.loads(control_setting_name.data)
    res = remote_controller.get_control(args)
    return pickle.dumps(res)


def rpc_server_main(sim_mode, serve_event, addr=default_addr, port=default_port):
    logger = init_logger(__name__)
    logger.info('controller process init')
    global remote_controller
    if addr != default_addr:
        raise NotImplementedError
    if port != default_port:
        raise NotImplementedError
    remote_controller = vent.controller.control_module.get_control_module(sim_mode)
    server = SimpleXMLRPCServer((addr, port), allow_none=True, logRequests=False)
    server.register_function(get_sensors, "get_sensors")
    # server.register_function(get_active_alarms, "get_active_alarms")
    # server.register_function(get_logged_alarms, "get_logged_alarms")
    server.register_function(set_control, "set_control")
    server.register_function(get_control, "get_control")
    server.register_function(remote_controller.start, "start")
    server.register_function(remote_controller.is_running, "is_running")
    server.register_function(remote_controller.stop, "stop")
    serve_event.set()
    server.serve_forever()



def get_rpc_client():
    # https://mail.python.org/pipermail/python-bugs-list/2015-January/260126.html
    #transport = xmlrpc.client.Transport()
    #con = transport.make_connection(f"http://{default_addr}:{default_port}/")
    #con.timeout = 5
    #
    proxy = xmlrpc.client.ServerProxy(f"http://{default_addr}:{default_port}/")

    return proxy


import pickle
import threading
from typing import List, Dict

import vent
import vent.controller.control_module
from vent.common.message import ControlSetting
from vent.alarm import Alarm
from vent.common.message import SensorValues
from vent.common.values import ValueName
from vent.common.logging import init_logger
from vent.coordinator.process_manager import ProcessManager
from vent.coordinator.rpc import get_rpc_client



class CoordinatorBase:
    def __init__(self, sim_mode=False):
        # get_ui_control_module handles single_process flag
        # self.lock = threading.Lock()
        self.logger = init_logger(__name__)
        self.logger.info('coordinator init')

    # TODO: do we still need this
    # def get_msg_timestamp(self):
    #     # return timestamp of last message
    #     with self.lock:
    #         last_message_timestamp = self.last_message_timestamp
    #     return last_message_timestamp


    def get_sensors(self) -> SensorValues:
        pass

    # def get_active_alarms(self) -> Dict[str, Alarm]:
    #     pass
    #
    # def get_logged_alarms(self) -> List[Alarm]:
    #     pass
    #
    # def clear_logged_alarms(self):
    #     pass

    def set_control(self, control_setting: ControlSetting):
        pass

    def get_control(self, control_setting_name: ValueName) -> ControlSetting:
        pass

    def start(self):
        pass

    def is_running(self) -> bool:
        pass

    def stop(self):
        pass

class CoordinatorLocal(CoordinatorBase):
    def __init__(self, sim_mode=False):
        """

        Args:
            sim_mode:

        Attributes:
            _is_running (:class:`threading.Event`): ``.set()`` when thread should stop

        """
        super().__init__(sim_mode=sim_mode)
        self.control_module = vent.controller.control_module.get_control_module(sim_mode)


    def get_sensors(self) -> SensorValues:

        # return res
        return self.control_module.get_sensors()

    # def get_active_alarms(self) -> Dict[str, Alarm]:
    #     return self.control_module.get_active_alarms()

    # def get_logged_alarms(self) -> List[Alarm]:
    #     return self.control_module.get_logged_alarms()

    # def clear_logged_alarms(self):
    #     # TODO: implement this
    #     raise NotImplementedError

    def set_control(self, control_setting: ControlSetting):
        self.control_module.set_control(control_setting)

    def get_control(self, control_setting_name: ValueName) -> ControlSetting:
        return self.control_module.get_control(control_setting_name)

    def start(self):
        """
        Start the coordinator.
        This does a soft start (not allocating a process).
        """
        self.control_module.start()

    def is_running(self) -> bool:
        """
        Test whether the whole system is running
        """
        return self.control_module._running

    def stop(self):
        """
        Stop the coordinator.
        This does a soft stop (not kill a process)
        """
        self.control_module.stop()


class CoordinatorRemote(CoordinatorBase):
    def __init__(self, sim_mode=False):
        super().__init__(sim_mode=sim_mode)
        # TODO: according to documentation, pass max_heartbeat_interval?
        self.process_manager = ProcessManager(sim_mode)
        self.rpc_client = get_rpc_client()
        # TODO: make sure the ipc connection is setup. There should be a clever method

    def get_sensors(self) -> SensorValues:
        sensor_values = pickle.loads(self.rpc_client.get_sensors().data)
        return sensor_values

    # def get_active_alarms(self) -> Dict[str, Alarm]:
    #     pickled_res = self.rpc_client.get_active_alarms().data
    #     return pickle.loads(pickled_res)
    #
    # def get_logged_alarms(self) -> List[Alarm]:
    #     pickled_res = self.rpc_client.get_logged_alarms().data
    #     return pickle.loads(pickled_res)
    #
    # def clear_logged_alarms(self):
    #     # TODO: implement this
    #     raise NotImplementedError

    def set_control(self, control_setting: ControlSetting):
        pickled_args = pickle.dumps(control_setting)
        self.rpc_client.set_control(pickled_args)

    def get_control(self, control_setting_name: ValueName) -> ControlSetting:
        pickled_args = pickle.dumps(control_setting_name)
        pickled_res = self.rpc_client.get_control(pickled_args).data
        return pickle.loads(pickled_res)

    def start(self):
        """
        Start the coordinator.
        This does a soft start (not allocating a process).
        """
        self.rpc_client.start()

    def is_running(self) -> bool:
        """
        Test whether the whole system is running
        """
        return self.rpc_client.is_running()

    def stop(self):
        """
        Stop the coordinator.
        This does a soft stop (not kill a process)
        """
        try:
            self.rpc_client.stop()
        except ConnectionRefusedError:
            pass
        self.process_manager.try_stop_process()

    def __del__(self):
        self.stop()


def get_coordinator(single_process=False, sim_mode=False) -> CoordinatorBase:
    if single_process:
        return CoordinatorLocal(sim_mode)
    else:
        return CoordinatorRemote(sim_mode)



import copy
import time

from vent.alarm import AlarmSeverity, AlarmType
from vent.alarm.condition import Condition
from vent.common.message import SensorValues, ControlSetting
from vent.alarm.alarm import Alarm
from vent.alarm.rule import Alarm_Rule

import typing

ALARM_MANAGER_INSTANCE = None

class Alarm_Manager(object):
    """
    Attributes:
        active_alarms (dict): {:class:`.AlarmType`: :class:`.Alarm`}
        pending_clears (list): [:class:`.AlarmType`] list of alarms that have been requested to be cleared
        callbacks (list): list of callables that accept `Alarm` s when they are raised/altered.
        cleared_alarms (list): of :class:`.AlarmType` s, alarms that have been cleared but have not dropped back into the 'off' range to enable re-raising
        snoozed_alarms (dict): of :class:`.AlarmType` s : times, alarms that should not be raised because they have been silenced for a period of time
    """
    _instance = None

    # use class attributes because __init__ is called every time instantiated

    active_alarms: typing.Dict[AlarmType, Alarm] = {}
    logged_alarms: typing.List[Alarm] = []

    # get our alarm rules
    dependencies = {}
    pending_clears = []
    cleared_alarms = []
    snoozed_alarms = {}
    callbacks = []
    rules = {}

    def __new__(cls):
        """
        If an Alarm_Manager already exists, when initing just return that one
        """

        if not cls._instance:
            cls._instance = super(Alarm_Manager, cls).__new__(cls)

        return cls._instance


    def __init__(self):

        if len(self.rules) == 0:
            self.load_rules()

    def load_rules(self):
        from vent.alarm import ALARM_RULES
        rules = copy.deepcopy(ALARM_RULES)

        # register dependencies
        for alarm_name, alarm_rule in rules.items():
            self.load_rule(alarm_rule)


    def load_rule(self, alarm_rule: Alarm_Rule):
        self.rules[alarm_rule.name] = alarm_rule

        for severity, condition in alarm_rule.conditions:

            if isinstance(condition.depends, dict):
                self.register_dependency(condition, condition.depends)
            elif isinstance(condition.depends, list) or isinstance(condition.depends, tuple):
                for depend in condition.depends:
                    self.register_dependency(condition, depend)



    def update(self, sensor_values: SensorValues):
        for alarm_name, rule in self.rules.items():
            self.check_rule(rule, sensor_values)
            # don't want to do alarm emission here because any _check_,
            # not any full update should trigger an alarm

    def check_rule(self, rule: Alarm_Rule, sensor_values: SensorValues):
        current_severity = rule.check(sensor_values)

        ##################
        # Checks that prevent raising

        # check that we're not being snoozed
        if rule.name in self.snoozed_alarms.keys():
            if time.time() >= self.snoozed_alarms[rule.name]:
                # remove from dict and continue
                del self.snoozed_alarms[rule.name]
            else:
                return

        # if we've been cleared, check if we've gone below zero,
        # otherwise return (don't re-raise)
        if rule.name in self.cleared_alarms:
            if current_severity == AlarmSeverity.OFF:
                self.cleared_alarms.remove(rule.name)
            else:
                return

        #####################
        # checks that determine type of raise

        if rule.name in self.active_alarms.keys():
            # if we've got an active alarm of this type
            # check if the severity has changed
            if current_severity.value > self.active_alarms[rule.name].severity.value:
                # greater severity always raises
                self.emit_alarm(rule.name, current_severity)
            elif current_severity.value < self.active_alarms[rule.name].severity.value:
                # if alarm isn't latched, emit lower alarm
                if not rule.latch:
                    self.emit_alarm(rule.name, current_severity)
                else:
                    # if the alarm is latched, but has been previously requested to be cleared...
                    if rule.name in self.pending_clears:
                        # clear if the severity is zero
                        if current_severity == AlarmSeverity.OFF:
                            self.emit_alarm(rule.name, current_severity)
                            self.pending_clears.remove(rule.name)
        else:
            # alarm isn't active
            if current_severity.value > AlarmSeverity.OFF.value:
                # emit if not off
                self.emit_alarm(rule.name, current_severity)



    def emit_alarm(self, alarm_type: AlarmType, severity: AlarmSeverity):
        """
        Emit alarm (by calling all callbacks with it).


        .. note::

            This method emits *and* clears alarms -- a cleared alarm is emitted with :attr:`AlarmSeverity.OFF`

        Args:
            alarm_type (:class:`.AlarmType`):
            severity (:class:`.AlarmSeverity`):
        """
        if alarm_type in self.rules.keys():
            # if another alarm is currently active, deactivate it
            if alarm_type in self.active_alarms.keys():
                self.deactivate_alarm(alarm_type)

            # make alarm and emit
            new_alarm = Alarm(
                alarm_type = alarm_type,
                severity   = severity,
                start_time = time.time(),
                latch      = self.rules[alarm_type].latch,
                persistent = self.rules[alarm_type].persistent,
                managed = True
            )

            for callback in self.callbacks:
                callback(new_alarm)

            if severity.value > AlarmSeverity.OFF.value:
                # don't add OFF alarms to active_alarms...
                self.active_alarms[alarm_type] = new_alarm

        else:
            raise ValueError('No  rule found for alarm type {}'.format(alarm_type))

    def deactivate_alarm(self, alarm: (AlarmType, Alarm)):
        """
        Mark an alarm's internal active flags and remove from :attr:`.active_alarms`

        .. note::

            This does *not* alert listeners that an alarm has been cleared,
            for that emit an alarm with AlarmSeverity.OFF

        Args:
            alarm:

        Returns:

        """

        if isinstance(alarm, Alarm):
            alarm_type = alarm.alarm_type

        elif isinstance(alarm, AlarmType):
            alarm_type = alarm

        else:
            raise ValueError(f'alarm must be AlarmType or Alarm, got {alarm}')

        if alarm_type in self.active_alarms.keys():
            if isinstance(alarm, Alarm):
                if alarm is not self.active_alarms[alarm_type]:
                    # if we were passed an Alarm and
                    # if this alarm isn't the one that's active, don't deactivate
                    return
            got_alarm = self.active_alarms.pop(alarm_type)
            got_alarm.deactivate()
            self.logged_alarms.append(got_alarm)
        else:
            return

    def dismiss_alarm(self,
                      alarm_type: AlarmType,
                      duration: float = None,):
        """
        GUI or other object requests an alarm to be dismissed & deactivated

        GUI will wait until it receives an `emit_alarm` of severity == OFF to remove
        alarm widgets. If the alarm is not latched

        If the alarm is latched, alarm_manager will not decrement alarm severity or emit
        `OFF` until a) the condition returns to `OFF`, and b) the user dismisses the alarm

        Args:
            alarm_type (:class:`.AlarmType`): Alarm to dismiss
            duration (float): seconds - amount of time to wait before alarm can be re-raised
                If a duration is provided, the alarm will not be able to be re-raised
        """

        assert(alarm_type in self.rules.keys())

        rule = self.rules[alarm_type]
        # if the alarm is latched, add it to the list of pending_clears
        if rule.latch:
            # if the alarm is in the pending_clears list,
            # when the `rule` returns to OFF, the alarm will be deactivated
            self.pending_clears.append(alarm_type)
            # the rest of the logic doesn't apply to latched alarms
            return

        # if we were provided a snooze duration, add to the snooze dict and clear alarm
        if duration:
            assert isinstance(duration, (int, float))
            assert duration >= 0
            self.snoozed_alarms[alarm_type] = time.time() + duration
            self.emit_alarm(alarm_type, AlarmSeverity.OFF)

        #otherwise alarm will be deactivated until condition goes OFF and back on
        else:
            self.cleared_alarms.append(alarm_type)
            self.emit_alarm(alarm_type, AlarmSeverity.OFF)

    def get_alarm_severity(self, alarm_type: AlarmType):
        if alarm_type in self.active_alarms.keys():
            return self.active_alarms[alarm_type].severity
        else:
            return AlarmSeverity.OFF


    def register_alarm(self, alarm: Alarm):
        """
        Add alarm to registry.

        Args:
            alarm (:class:`.Alarm`)
        """

        if alarm.alarm_type in self.active_alarms.keys():
            if alarm is self.active_alarms[alarm.alarm_type]:
                return
            # if another alarm is already active,
            # check if this is a higher severity
            if alarm.severity.value > self.active_alarms[alarm.alarm_type].severity.value:
                self.deactivate_alarm(alarm.alarm_type)
                self.active_alarms[alarm.alarm_type] = alarm
            else:
                return
                # TODO: currently just bouncing redundant alarms, is that what we want?
        else:
            self.active_alarms[alarm.alarm_type] = alarm

        # TODO: Emit evidence of this new alarm

    def register_dependency(self, condition: Condition,
                            dependency: dict):
        """
        Add dependency in a Condition object to be updated when values are changed

        Args:
            condition:
            dependency (dict): either a (ValueName, attribute_name) or optionally also + transformation callable
        """

        # invert the structure of the dependency so it's keyed by the ValueName being updated
        dependency['condition'] = condition,
        value_name = dependency.pop('value_name')
        if value_name not in self.dependencies.keys():
            self.dependencies[value_name] = [dependency]
        else:
            self.dependencies[value_name].append(dependency)


    def update_dependencies(self, control_setting: ControlSetting):
        """
        Update Condition objects that update their value according to some control parameter

        Args:
            control_setting (:class:`.ControlSetting`):

        Returns:

        """
        if control_setting.name in self.dependencies.keys():
            # dependencies have
            # ('value_name', 'value_attribute', 'condition_attr', 'transform')
            for depend in self.dependencies[control_setting.name]:

                # get the value from the control setting we want to update from
                new_value = getattr(control_setting, depend['value_attribute'])

                # if the attribute we're looking to update from is None, then
                # this control isn't being set right now
                if new_value is None:
                    return

                if 'transform' in depend.keys():
                    # apply transformation if it has one
                    new_value = depend['transform'](new_value)

                setattr(depend['condition'], depend['condition_attr'], new_value)

    def add_callback(self, callback: typing.Callable):
        assert callable(callback)
        self.callbacks.append(callback)

    def clear_all_alarms(self):
        # make separate list because dict will be cleared during iteration
        alarm_keys = list(self.active_alarms.keys())
        for alarm_type in alarm_keys:
            self.deactivate_alarm(alarm_type)

    def reset(self):
        """
        reset all conditions, callbacks, and other stateful attributes and clear alarms
        """

        self.pending_clears = []
        self.cleared_alarms = []
        self.snoozed_alarms = {}
        self.callbacks = []

        for rule in self.rules.values():
            rule.reset()

        self.clear_all_alarms()












import time
from typing import List
import threading
import numpy as np
import copy
from collections import deque
import pdb
from itertools import count

import vent.io as io

from vent.common.message import SensorValues, ControlSetting
from vent.common.logging import init_logger
from vent.common.values import CONTROL, ValueName
from vent.alarm import AlarmSeverity, Alarm



class ControlModuleBase:
    """This is an abstract class for controlling simulation and hardware.

    1. All internal variables fall in three classes, denoted by the beginning of the variable:
        - "COPY_varname": These are copies (see 1.) that are regularly sync'ed with internal variables.
        - "__varname":    These are variables only used in the ControlModuleBase-Class
        - "_varname":     These are variables used in derived classes.

    2. Internal variables should only to be accessed though the set_ and get_ functions.
        These functions act on COPIES of internal variables ("__" and "_"), that are sync'd every few
        iterations. How often this is done is adjusted by the variable
        self._NUMBER_CONTROLL_LOOPS_UNTIL_UPDATE. To avoid multiple threads manipulating the same 
        variables at the same time, every manipulation of "COPY_" is surrounded by a thread lock.

    Public Methods:
        get_sensors():                     Returns a copy of the current sensor values.
        get_alarms():                      Returns a List of all alarms, active and logged
        get_active_alarms():               Returns a Dictionary of all currently active alarms.
        get_logged_alarms():               Returns a List of logged alarms, up to maximum lengh of self._RINGBUFFER_SIZE
        get_control(ControlSetting):       Sets a controll-setting. Is updated at latest within self._NUMBER_CONTROLL_LOOPS_UNTIL_UPDATE
        get_past_waveforms():              Returns a List of waveforms of pressure and volume during at the last N breath cycles, N<self._RINGBUFFER_SIZE, AND clears this archive.
        get_target_waveform():             Returns a step-wise linear target waveform, as defined by the current settings.
        start():                           Starts the main-loop of the controller
        stop():                            Stops the main-loop of the controller

    """

    def __init__(self):
        self.logger = init_logger(__name__)
        self.logger.info('controller init')
        #####################  Algorithm/Program parameters  ##################
        # Hyper-Parameters
        self._LOOP_UPDATE_TIME                   = 0.01    # Run the main control loop every 0.01 sec
        self._NUMBER_CONTROLL_LOOPS_UNTIL_UPDATE = 10      # After every 10 main control loop iterations, update COPYs.
        self._RINGBUFFER_SIZE                    = 100     # Maximum number of breath cycles kept in memory

        #########################  Control management  #########################

        # This is what the machine has controll over:
        self.__control_signal_in  = 0              # State of a valve on the inspiratory side - could be a proportional valve.
        self.__control_signal_out = 0              # State of a valve on the exspiratory side - this is open/close i.e. value in (0,1)
        self._pid_control_flag    = True           # Default is: use PID control
        self.__KP                 = 80             # The weights for the the PID terms
        self.__KI                 = 0
        self.__KD                 = 0


        # Internal Control variables. "SET" indicates that this is set.
        self.__SET_PIP       = CONTROL[ValueName.PIP].default                     # Target PIP pressure
        self.__SET_PIP_TIME  = CONTROL[ValueName.PIP_TIME].default                # Target time to reach PIP in seconds
        self.__SET_PEEP      = CONTROL[ValueName.PEEP].default                    # Target PEEP pressure
        self.__SET_PEEP_TIME = CONTROL[ValueName.PEEP_TIME].default               # Target time to reach PEEP from PIP plateau
        self.__SET_BPM       = CONTROL[ValueName.BREATHS_PER_MINUTE].default      # Target breaths per minute
        self.__SET_I_PHASE   = CONTROL[ValueName.INSPIRATION_TIME_SEC].default    # Target duration of inspiratory phase

        # Derived internal control variables - fully defined by numbers above
        self.__SET_CYCLE_DURATION = 60 / self.__SET_BPM
        self.__SET_E_PHASE        = self.__SET_CYCLE_DURATION - self.__SET_I_PHASE
        self.__SET_T_PLATEAU      = self.__SET_I_PHASE - self.__SET_PIP_TIME
        self.__SET_T_PEEP         = self.__SET_E_PHASE - self.__SET_PEEP_TIME

        #########################  Data management  #########################

        # These are measurements from the last breath cycle.
        self._DATA_PIP = None         # Measured value of PIP
        self._DATA_PIP_PLATEAU = None # Measured pressure of the plateau
        self._DATA_PIP_TIME = None    # Measured time of reaching PIP plateau
        self._DATA_PEEP = None        # Measured valued of PEEP
        self._DATA_I_PHASE = None     # Measured duration of inspiratory phase
        self.__DATA_LAST_PEEP = None  # Last time of PEEP - by definition end of breath cycle
        self._DATA_BPM = None         # Measured breathing rate, by definition 60sec / length_of_breath_cycle
        self._DATA_VTE = None         # Maximum air displacement in last breath cycle
        self._DATA_P = 0              # Last measurements of the proportional term for PID-control
        self._DATA_I = 0              # Last measurements of the integral term for PID-control
        self._DATA_D = 0              # Last measurements of the differential term for PID-control
        self._DATA_BREATH_COUNT = 0   # Total number of breaths/id of current breath cycle
        self._breath_counter = count() # threadsafe counter

        # Parameters to keep track of breath-cycle
        self.__cycle_start = time.time()
        self.__cycle_waveform = np.array([[0, 0, 0]])                            # To build up the current cycle's waveform
        self.__cycle_waveform_archive = deque(maxlen = self._RINGBUFFER_SIZE)          # An archive of past waveforms.

        # These are measurements that change from timepoint to timepoint
        self._DATA_PRESSURE = 0
        self.__DATA_VOLUME  = 0
        self._DATA_Qin      = 0           # Measurement of the airflow in
        self._DATA_Qout     = 0           # Measurement of the airflow out
        self._DATA_dpdt     = 0           # Current sample of the rate of change of pressure dP/dt in cmH2O/sec
        self._last_update   = time.time()


        #########################  Alarm management  #########################
        self.__active_alarms = {}     # Dictionary of active alarms
        self.__logged_alarms = deque(maxlen = self._RINGBUFFER_SIZE)     # List of all resolved alarms

        # Variable limits to raise alarms, initialized as small deviation of what the controller initializes
        self.__PIP_min          = CONTROL[ValueName.PIP].safe_range[0]
        self.__PIP_max          = CONTROL[ValueName.PIP].safe_range[1]
        self.__PIP_lastset      = time.time()
        self.__PIP_time_min     = CONTROL[ValueName.PIP_TIME].safe_range[0]
        self.__PIP_time_max     = CONTROL[ValueName.PIP_TIME].safe_range[1]
        self.__PIP_time_lastset = time.time()
        self.__PEEP_min         = CONTROL[ValueName.PEEP].safe_range[0]
        self.__PEEP_max         = CONTROL[ValueName.PEEP].safe_range[1]
        self.__PEEP_lastset     = time.time()
        self.__PEEP_time_min    = CONTROL[ValueName.PEEP_TIME].safe_range[0]
        self.__PEEP_time_max    = CONTROL[ValueName.PEEP_TIME].safe_range[1]
        self.__PEEP_time_lastset = time.time()
        self.__bpm_min          = CONTROL[ValueName.BREATHS_PER_MINUTE].safe_range[0]
        self.__bpm_max          = CONTROL[ValueName.BREATHS_PER_MINUTE].safe_range[1]
        self.__bpm_lastset      = time.time()
        self.__I_phase_min      = CONTROL[ValueName.INSPIRATION_TIME_SEC].safe_range[0]
        self.__I_phase_max      = CONTROL[ValueName.INSPIRATION_TIME_SEC].safe_range[1]
        self.__I_phase_lastset  = time.time()

        ############### Initialize COPY variables for threads  ##############
        # COPY variables that later updated on a regular basis
        self.COPY_active_alarms = {}
        self.COPY_logged_alarms = list(self.__logged_alarms)
        self.COPY_sensor_values = None # empty SensorValues can no longer be instantiated -jls

        ###########################  Threading init  #########################
        # Run the start() method as a thread
        self._loop_counter = 0
        self._running = threading.Event()
        self._running.clear()
        self._lock = threading.Lock()
        self._alarm_to_COPY()  #These require the lock
        self._initialize_set_to_COPY()

        # self.__thread = threading.Thread(target=self._start_mainloop, daemon=True)
        # self.__thread.start()
        self.__thread = None


    def _initialize_set_to_COPY(self):
        self._lock.acquire()
        # Copy of the SET variables for threading.
        self.COPY_SET_PIP       = self.__SET_PIP 
        self.COPY_SET_PIP_TIME  = self.__SET_PIP_TIME
        self.COPY_SET_PEEP      = self.__SET_PEEP
        self.COPY_SET_PEEP_TIME = self.__SET_PEEP_TIME
        self.COPY_SET_BPM       = self.__SET_BPM
        self.COPY_SET_I_PHASE   = self.__SET_I_PHASE
        self._lock.release()

    def _alarm_to_COPY(self):
        self._lock.acquire()
        # Update the alarms
        self.COPY_active_alarms = self.__active_alarms.copy()
        self.COPY_logged_alarms = list(self.__logged_alarms)

        # The alarm thresholds
        self.COPY_PIP_min = self.__PIP_min
        self.COPY_PIP_max = self.__PIP_max
        self.COPY_PIP_lastset = self.__PIP_lastset
        self.COPY_PIP_time_min = self.__PIP_time_min 
        self.COPY_PIP_time_max = self.__PIP_time_max
        self.COPY_PIP_time_lastset = self.__PIP_time_lastset
        self.COPY_PEEP_min = self.__PEEP_min
        self.COPY_PEEP_max = self.__PEEP_max
        self.COPY_PEEP_lastset = self.__PEEP_lastset
        self.COPY_PEEP_time_min = self.__PEEP_time_min
        self.COPY_PEEP_time_max = self.__PEEP_time_max
        self.COPY_PEEP_time_lastset = self.__PEEP_time_lastset
        self.COPY_bpm_min = self.__bpm_min 
        self.COPY_bpm_max = self.__bpm_max
        self.COPY_bpm_lastset = self.__bpm_lastset
        self.COPY_I_phase_min = self.__I_phase_min
        self.COPY_I_phase_max = self.__I_phase_max
        self.COPY_I_phase_lastset = self.__I_phase_lastset

        self._lock.release()

    def _sensor_to_COPY(self):
        # These variables have to come from the hardware
        self._lock.acquire()
        # Make sure you have acquire and release!
        self._lock.release()
        pass

    def _controls_from_COPY(self):
        # Update SET variables
        self._lock.acquire()

        #Update values
        self.__SET_PIP       = self.COPY_SET_PIP
        self.__SET_PIP_TIME  = self.COPY_SET_PIP_TIME
        self.__SET_PEEP      = self.COPY_SET_PEEP
        self.__SET_PEEP_TIME = self.COPY_SET_PEEP_TIME
        self.__SET_BPM       = self.COPY_SET_BPM
        self.__SET_I_PHASE   = self.COPY_SET_I_PHASE

        #Update derived values
        self.__SET_CYCLE_DURATION = 60 / self.__SET_BPM
        self.__SET_E_PHASE = self.__SET_CYCLE_DURATION - self.__SET_I_PHASE
        self.__SET_T_PLATEAU = self.__SET_I_PHASE - self.__SET_PIP_TIME
        self.__SET_T_PEEP = self.__SET_E_PHASE - self.__SET_PEEP_TIME

        #Update new confidence intervals
        self.__PIP_min          = self.COPY_PIP_min
        self.__PIP_max          = self.COPY_PIP_max 
        self.__PIP_lastset      = self.COPY_PIP_lastset  
        self.__PIP_time_min     = self.COPY_PIP_time_min  
        self.__PIP_time_max     = self.COPY_PIP_time_max  
        self.__PIP_time_lastset = self.COPY_PIP_time_lastset  
        self.__PEEP_min         = self.COPY_PEEP_min 
        self.__PEEP_max         = self.COPY_PEEP_max  
        self.__PEEP_lastset     = self.COPY_PEEP_lastset
        self.__PEEP_time_min    = self.COPY_PEEP_time_min
        self.__PEEP_time_max    = self.COPY_PEEP_time_max
        self.__PEEP_time_lastset = self.COPY_PEEP_time_lastset
        self.__bpm_min          = self.COPY_bpm_min 
        self.__bpm_max          = self.COPY_bpm_max  
        self.__bpm_lastset      = self.COPY_bpm_lastset  
        self.__I_phase_min      = self.COPY_I_phase_min  
        self.__I_phase_max      = self.COPY_I_phase_max 
        self.__I_phase_lastset  = self.COPY_I_phase_lastset  

        self._lock.release()

    def __test_critical_levels(self, min, max, value, name):
        '''
        This tests whether a variable is within bounds.
        If it is, and an alarm existed, then the "alarm_end_time" is set.
        If it is NOT, a new alarm is generated and appendede to the alarm-list.
        Input:
            min:           minimum value  (e.g. 2)
            max:           maximum value  (e.g. 5)
            value:         test value   (e.g. 3)
            name:          parameter type (e.g. "PIP", "PEEP" etc.)
        '''
        pass
        # if (value < min) or (value > max):  # If the variable is not within limits
        #     if name not in self.__active_alarms.keys():  # And and alarm for that variable doesn't exist yet -> RAISE ALARM.
        #         new_alarm = Alarm(alarm_name=name, active=True, severity=AlarmSeverity.HIGH, value=value,
        #                           start_time=time.time(), alarm_end_time=None)
        #         self.__active_alarms[name] = new_alarm
        # else:  # Else: if the variable is within bounds,
        #     if name in self.__active_alarms.keys():  # And an alarm exists -> inactivate it.
        #         old_alarm = self.__active_alarms[name]
        #         old_alarm.alarm_end_time = time.time()
        #         old_alarm.active = False
        #         self.__logged_alarms.append(old_alarm)
        #         del self.__active_alarms[name]

    def __analyze_last_waveform(self):
        ''' This goes through the last waveform, and updates VTE, PEEP, PIP, PIP_TIME, I_PHASE, FIRST_PEEP and BPM.'''
        if len(self.__cycle_waveform_archive) > 1:  # Only if there was a previous cycle
            data = self.__cycle_waveform_archive[-1]
            phase = data[:, 0]
            pressure = data[:, 1]
            mean_pressure = np.mean(pressure)
            volume = data[:, 2]

            self._DATA_VTE = np.max(volume) - np.min(volume)

            # get the pressure niveau heuristically (much faster than fitting)
            # 20/80 percentile of pressure values below/above mean
            # Assumption: waveform is mostly between both plateaus
            if np.isfinite(mean_pressure):
                self._DATA_PEEP = np.percentile(pressure[ pressure < mean_pressure], 20 )
                self._DATA_PIP_PLATEAU  = np.percentile(pressure[ pressure > mean_pressure], 80 )
                self._DATA_PIP  = np.percentile(pressure[ pressure > mean_pressure], 95 )             #PIP is defined as the maximum, here 95% to account for outliers
                self._DATA_PIP_TIME = phase[np.min(np.where(pressure > self._DATA_PIP_PLATEAU*0.9))]
                self._DATA_I_PHASE = phase[np.max(np.where(pressure > self._DATA_PIP_PLATEAU*0.9))]
            else:
                self._DATA_PEEP = np.nan
                self._DATA_PIP_PLATEAU  = np.nan
                self._DATA_PIP  = np.nan
                self._DATA_PIP_TIME = np.nan
                self._DATA_I_PHASE = np.nan

            # and measure the breaths per minute
            self._DATA_BPM = 60. / phase[-1]  # 60 sec divided by the duration of last waveform

    def __update_alarms(self):
        ''' This goes through the values obtained from the last waveform, and updates alarms.'''
        if len(self.__cycle_waveform_archive) > 1 : # Only if there was a previous cycle
            self.__test_critical_levels(min=self.__PIP_min, max=self.__PIP_max, value=self._DATA_PIP, name=ValueName.PIP)
            self.__test_critical_levels(min=self.__PIP_time_min, max=self.__PIP_time_max, value=self._DATA_PIP_TIME, name=ValueName.PIP_TIME)
            self.__test_critical_levels(min=self.__PEEP_min, max=self.__PEEP_max, value=self._DATA_PEEP, name=ValueName.PEEP)
            self.__test_critical_levels(min=self.__bpm_min, max=self.__bpm_max, value=self._DATA_BPM, name=ValueName.BREATHS_PER_MINUTE)
            self.__test_critical_levels(min=self.__I_phase_min, max=self.__I_phase_max, value=self._DATA_I_PHASE, name=ValueName.INSPIRATION_TIME_SEC)

    def get_sensors(self) -> SensorValues:
        # Make sure to return a copy of the instance
        self._lock.acquire()
        #cp = copy.deepcopy( self.COPY_sensor_values )
        # don't need to deepcopy because a new SensorValues object is created
        # each time and it copies each individual value as it's created
        cp = copy.copy(self.COPY_sensor_values)
        self._lock.release()
        return cp

    # def get_alarms(self) -> List[Alarm]:
    #     # Returns all alarms as a list
    #     self._lock.acquire()
    #     new_alarm_list = self.COPY_logged_alarms.copy()
    #     for alarm_key in self.COPY_active_alarms.keys():
    #         new_alarm_list.append(self.COPY_active_alarms[alarm_key])
    #     self._lock.release()
    #     return new_alarm_list

    def get_active_alarms(self):
        # Returns only the active alarms
        self._lock.acquire()
        active_alarms = self.COPY_active_alarms.copy() # Make sure to return a copy
        self._lock.release()
        return active_alarms

    def get_logged_alarms(self) -> List[Alarm]:
        # Returns only the inactive alarms
        self._lock.acquire()
        logged_alarms = self.COPY_logged_alarms.copy()
        self._lock.release()
        return logged_alarms



    def set_control(self, control_setting: ControlSetting):
        ''' Updates the entry of COPY contained in the control settings'''
        self._lock.acquire()

        if control_setting.name == ValueName.PIP:
            self.COPY_SET_PIP = control_setting.value
            self.COPY_PIP_min = control_setting.min_value
            self.COPY_PIP_max = control_setting.max_value
            self.COPY_PIP_lastset = control_setting.timestamp

        elif control_setting.name == ValueName.PIP_TIME:
            self.COPY_SET_PIP_TIME = control_setting.value
            self.COPY_PIP_time_min = control_setting.min_value
            self.COPY_PIP_time_max = control_setting.max_value
            self.COPY_PIP_time_lastset = control_setting.timestamp

        elif control_setting.name == ValueName.PEEP:
            self.COPY_SET_PEEP = control_setting.value
            self.COPY_PEEP_min = control_setting.min_value
            self.COPY_PEEP_max = control_setting.max_value
            self.COPY_PEEP_lastset = control_setting.timestamp

        elif control_setting.name == ValueName.BREATHS_PER_MINUTE:
            self.COPY_SET_BPM = control_setting.value
            self.COPY_bpm_min = control_setting.min_value
            self.COPY_bpm_max = control_setting.max_value
            self.COPY_bpm_lastset = control_setting.timestamp

        elif control_setting.name == ValueName.INSPIRATION_TIME_SEC:
            self.COPY_SET_I_PHASE = control_setting.value
            self.COPY_I_phase_min = control_setting.min_value
            self.COPY_I_phase_max = control_setting.max_value
            self.COPY_I_phase_lastset = control_setting.timestamp

        elif control_setting.name == ValueName.PEEP_TIME:
            self.COPY_SET_PEEP_TIME = control_setting.value
            self.COPY_PEEP_min = control_setting.min_value
            self.COPY_PEEP_max = control_setting.max_value
            self.COPY_PEEP_lastset = control_setting.timestamp

        else:
            raise KeyError("You cannot set the variabe: " + str(control_setting.name))

        self._lock.release()

    def get_control(self, control_setting_name: ValueName) -> ControlSetting:
        ''' Gets values of the COPY of the control settings. '''
        
        self._lock.acquire()
        if control_setting_name == ValueName.PIP:
            return_value = ControlSetting(control_setting_name,
                                  self.COPY_SET_PIP,
                                  self.COPY_PIP_min,
                                  self.COPY_PIP_max,
                                  self.COPY_PIP_lastset)
        elif control_setting_name == ValueName.PIP_TIME:
            return_value = ControlSetting(control_setting_name,
                                  self.COPY_SET_PIP_TIME,
                                  self.COPY_PIP_time_min,
                                  self.COPY_PIP_time_max,
                                  self.COPY_PIP_time_lastset, )
        elif control_setting_name == ValueName.PEEP:
            return_value = ControlSetting(control_setting_name,
                                  self.COPY_SET_PEEP,
                                  self.COPY_PEEP_min,
                                  self.COPY_PEEP_max,
                                  self.COPY_PEEP_lastset)
        elif control_setting_name == ValueName.BREATHS_PER_MINUTE:
            return_value = ControlSetting(control_setting_name,
                                  self.COPY_SET_BPM,
                                  self.COPY_bpm_min,
                                  self.COPY_bpm_max,
                                  self.COPY_bpm_lastset)
        elif control_setting_name == ValueName.INSPIRATION_TIME_SEC:
            return_value = ControlSetting(control_setting_name,
                                  self.COPY_SET_I_PHASE,
                                  self.COPY_I_phase_min,
                                  self.COPY_I_phase_max,
                                  self.COPY_I_phase_lastset)
        elif control_setting_name == ValueName.PEEP_TIME:
            return_value = ControlSetting(control_setting_name,
                                          self.COPY_SET_PEEP_TIME,
                                          self.COPY_PEEP_time_min,
                                          self.COPY_PEEP_time_max,
                                          self.COPY_PEEP_time_lastset)
        else:
            raise KeyError("You cannot set the variabe: " + str(control_setting_name))

        self._lock.release()

        return return_value


    def __get_PID_error(self, ytarget, yis, dt):
        error_new = ytarget - yis                   # New value of the error

        RC = 0.5  # Time constant in seconds
        s = dt / (dt + RC)
        self._DATA_I = self._DATA_I + s*(error_new - self._DATA_I)     # Integral term on some timescale RC
        self._DATA_D = error_new - self._DATA_P
        self._DATA_P = error_new

    def __calculate_control_signal_in(self):
        self.__control_signal_in  = 0            # Some setting for the maximum flow.
        self.__control_signal_in +=  self.__KP*self._DATA_P
        self.__control_signal_in +=  self.__KI*self._DATA_I
        self.__control_signal_in +=  self.__KD*self._DATA_D

    def _get_control_signal_in(self):
        ''' This is the PID controlled signal on the inspiratory side '''
        return self.__control_signal_in

    def _get_control_signal_out(self):
        ''' This is the control signal (open/close) on the expiratory side '''
        return self.__control_signal_out

    def _PID_reset(self):
        ''' resets the PID cycle to zero'''
        self.__cycle_start = time.time()

    def _PID_update(self, dt):
        ''' 
        This instantiates the control algorithms.
        During the breathing cycle, it goes through the four states:
           1) Rise to PIP
           2) Sustain PIP pressure
           3) Quick fall to PEEP
           4) Sustaint PEEP pressure
        Once the cycle is complete, it checks the cycle for any alarms, and starts a new one.
        A record of pressure/volume waveforms is kept in self.__cycle_waveform_archive

            dt: Time since last update in seconds 

        '''
        now = time.time()
        cycle_phase = now - self.__cycle_start
        next_cycle = False

        self.__DATA_VOLUME += dt * ( self._DATA_Qin - self._DATA_Qout )  # Integrate what has happened within the last few seconds from the measurements of Qin and Qout

        if cycle_phase < self.__SET_PIP_TIME:

            if self._pid_control_flag:
                target_pressure = cycle_phase*(self.__SET_PIP - self.__SET_PEEP) / self.__SET_PIP_TIME  + self.__SET_PEEP
                self.__get_PID_error(yis = self._DATA_PRESSURE, ytarget = target_pressure, dt = dt)
                self.__calculate_control_signal_in()
                self.__control_signal_out = 0   # close out valve
                if self._DATA_PRESSURE > self.__SET_PIP:
                    self.__control_signal_in = 0
            else:
                self.__control_signal_in = np.inf                                                        # STATE CONTROL: to PIP, air in as fast as possible
                self.__control_signal_out = 0
                if self._DATA_PRESSURE > self.__SET_PIP:
                    self.__control_signal_in = 0

        elif cycle_phase < self.__SET_I_PHASE:                                                           # then, we control PIP
            if self._pid_control_flag:
                self.__get_PID_error(yis = self._DATA_PRESSURE, ytarget = self.__SET_PIP, dt = dt)
                self.__calculate_control_signal_in()
                if self._DATA_PRESSURE > self.__SET_PIP+0.5:                                              
                    self.__control_signal_out = 1                                                        # if exceeded, we open the exhaust valve
                else:
                    self.__control_signal_out = 0                                                        # close out valve
            else:
                self.__control_signal_in = 0                                                             # STATE CONTROL: keep PIP plateau, let air in if below
                self.__control_signal_out = 0
                if self._DATA_PRESSURE < self.__SET_PIP:
                    self.__control_signal_in = np.inf
                if self._DATA_PRESSURE > self.__SET_PIP:
                    self.__control_signal_out = 1

        elif cycle_phase < self.__SET_PEEP_TIME + self.__SET_I_PHASE:                                     # then, we drop pressure to PEEP
            if self._pid_control_flag:
                target_pressure = self.__SET_PIP - (cycle_phase - self.__SET_I_PHASE) * (self.__SET_PIP - self.__SET_PEEP) / self.__SET_PEEP_TIME
                self.__get_PID_error(yis = self._DATA_PRESSURE, ytarget = target_pressure, dt = dt)
                self.__calculate_control_signal_in()
                self.__control_signal_out =  1
                if self._DATA_PRESSURE < self.__SET_PEEP - 1:
                    self.__control_signal_out = 0
                    self.__control_signal_in = 0
            else:
                self.__control_signal_in = 0
                self.__control_signal_out = 1
                if self._DATA_PRESSURE < self.__SET_PEEP:
                    self.__control_signal_out = 0
        elif cycle_phase < self.__SET_CYCLE_DURATION:                                                     # and control around PEEP
            if self._pid_control_flag:
                self.__get_PID_error(yis = self._DATA_PRESSURE, ytarget = self.__SET_PEEP, dt = dt)
                self.__calculate_control_signal_in()
                if self._DATA_PRESSURE > self.__SET_PEEP + 1:
                    self.__control_signal_out = 1
                else:
                    self.__control_signal_out = 0
            else:                                                                                         # STATE CONTROL: keeping PEEP, let air in if below
                self.__control_signal_in = 0
                self.__control_signal_out = 0
                if self._DATA_PRESSURE < self.__SET_PEEP:
                    self.__control_signal_in = np.inf
                if self._DATA_PRESSURE > self.__SET_PEEP:
                    self.__control_signal_out = 1

        else:
            self.__cycle_start = time.time()  # New cycle starts
            self.__DATA_VOLUME = 0            # ... start at zero volume in the lung
            self._DATA_dpdt    = 0            # and restart the rolling average for the dP/dt estimation
            next_cycle = True
        
        if next_cycle:                        # if a new breath cycle has started
            # increment breath_cycle tracker
            self._DATA_BREATH_COUNT = next(self._breath_counter)
            if len(self.__cycle_waveform) > 1:
                self.__cycle_waveform_archive.append( self.__cycle_waveform )
            self.__cycle_waveform = np.array([[0, self._DATA_PRESSURE, self.__DATA_VOLUME]])
            self.__analyze_last_waveform()    # Analyze last waveform
            self.__update_alarms()            # Run alarm detection over last cycle's waveform
            self._sensor_to_COPY()            # Get the fit values from the last waveform directly into sensor values
        else:
            self.__cycle_waveform = np.append(self.__cycle_waveform, [[cycle_phase, self._DATA_PRESSURE, self.__DATA_VOLUME]], axis=0)

    def get_past_waveforms(self):
        # Returns a list of past waveforms.
        # Format:
        #     Returns a list of [Nx3] waveforms, of [time, pressure, volume]
        #     Most recent entry is waveform_list[-1]
        # Note:
        #     After calling this function, archive is emptied!
        self._lock.acquire()
        archive = list( self.__cycle_waveform_archive ) # Make sure to return a copy as a list
        self.__cycle_waveform_archive = deque(maxlen = self._RINGBUFFER_SIZE)
        self.__cycle_waveform_archive.append(archive[-1])
        self._lock.release()
        return archive

    def get_target_waveform(self):
        # Returns the target waveform, drawn as a sketch of a stepwise linear function
        # Format is time-points, pressure values - to be connected with straight lines
        #         ______
        #        /      \                         <- Sketch waveform of single breath cycle
        #       /        \
        #      /          \____________
        #
        #     ^  ^     ^  ^           ^
        #     A  B     C  D           E           <- Critical time points

        self._lock.acquire()
        wv = (
        (0, self.__SET_PEEP),                                            # A: start of the waveform
        (self.__SET_PIP_TIME, self.__SET_PIP),                           # B: reaching PIP within PIP_TIME
        (self.__SET_I_PHASE, self.__SET_PIP),                            # C: keeping the plateau during I_Phase
        (self.__SET_PEEP_TIME + self.__SET_I_PHASE, self.__SET_PEEP),    # D: reaching PEEP within PEEP TIME
        (self.__SET_CYCLE_DURATION, self.__SET_PEEP))                    # E: Cycle ends
        self._lock.release()
        return wv

    def _start_mainloop(self):
        # This will depend on simulation or reality
        pass   

    def start(self):
        if self.__thread is None or not self.__thread.is_alive():  # If the previous thread has been stopped, make a new one.
            self._running.set()
            self.__thread = threading.Thread(target=self._start_mainloop, daemon=True)
            self.__thread.start()
        else:
            print("Main Loop already running.")

    def stop(self):
        if self.__thread is not None and self.__thread.is_alive():
            self._running.clear()
        else:
            print("Main Loop is not running.")

    def is_running(self):
        # TODO: this should be better thread-safe variable
        return self._running.is_set()

    def do_pid_control(self):
        if self._pid_control_flag:
            print("Already running PID control.")
        self._pid_control_flag = True

    def do_state_control(self):
        if not self._pid_control_flag:
            print("Already running State control.")
        self._pid_control_flag = False


class ControlModuleDevice(ControlModuleBase):
    # Implement ControlModuleBase functions
    def __init__(self):
        ControlModuleBase.__init__(self)
        self.HAL = io.Hal()
        self._sensor_to_COPY()
        
    def _sensor_to_COPY(self):
        # And the sensor measurements
        self._lock.acquire()
        self.COPY_sensor_values = SensorValues(vals={
            ValueName.PIP.name                  : self._DATA_PIP,
            ValueName.PEEP.name                 : self._DATA_PEEP,
            ValueName.FIO2.name                 : 70,
            ValueName.TEMP.name                 : -1,
            ValueName.HUMIDITY.name             : -1,
            ValueName.PRESSURE.name             : self.HAL.pressure,
            ValueName.VTE.name                  : self._DATA_VTE,
            ValueName.BREATHS_PER_MINUTE.name   : self._DATA_BPM,
            ValueName.INSPIRATION_TIME_SEC.name : self._DATA_I_PHASE,
            'timestamp'                  : time.time(),
            'loop_counter'             : self._loop_counter,
            'breath_count': self._DATA_BREATH_COUNT
        })
        self._lock.release()

    def _start_mainloop(self):
        # start running, this should be run as a thread! 
        # Compare to initialization in Base Class!

        update_copies = self._NUMBER_CONTROLL_LOOPS_UNTIL_UPDATE

        while self._running:
            time.sleep(self._LOOP_UPDATE_TIME)
            self._loop_counter += 1
            now = time.time()
            dt = now - self._last_update                            # Time sincle last cycle of main-loop

            if dt > CONTROL[ValueName.BREATHS_PER_MINUTE].default / 4:                                                      # TODO: RAISE HARDWARE ALARM, no update should be so long
                print("Restarted cycle.")
                self._PID_reset()
                dt = self._LOOP_UPDATE_TIME

            self._DATA_PRESSURE = self.HAL.pressure                 # Get a pressure measurement from HAL

            self._PID_update(dt = dt)                               # Update the PID Controller

            valve_open_in = self._get_control_signal_in()           # Inspiratory side: get control signal for PropValve
            self.HAL.setpoint_in = max(min(100, valve_open_in), 0)

            self.HAL.setpoint_ex = self._get_control_signal_out()          # Expiratory side: get control signal for Solenoid
            '''
            if(self.HAL.setpoint_ex == 0):
                self.HAL._expiratory_valve.close()
            else:
                self.HAL._expiratory_valve.open()
            '''

            self._DATA_Qout = self.HAL.flow_ex                     # Flow sensor on Expiratory side
            self._DATA_Qin  = self.HAL.flow_in                      # Flow sensor on inspiratory side. NOTE: used to calculate VTE
            self._last_update = now

            if update_copies == 0:
                self._controls_from_COPY()     # Update controls from possibly updated values as a chunk
                self._alarm_to_COPY()          # Copy current alarms and settings to COPY
                self._sensor_to_COPY()         # Copy sensor values to COPY
                update_copies = self._NUMBER_CONTROLL_LOOPS_UNTIL_UPDATE
            else:
                update_copies -= 1

        # # get final values on stop
        self._controls_from_COPY()  # Update controls from possibly updated values as a chunk
        self._alarm_to_COPY()  # Copy current alarms and settings to COPY
        self._sensor_to_COPY()  # Copy sensor values to COPY



class Balloon_Simulator:
    '''
    This is a imple physics simulator for inflating a balloon. 
    For math, see https://en.wikipedia.org/wiki/Two-balloon_experiment
    '''

    def __init__(self, leak):
        # Hard parameters for the simulation
        self.max_volume = 6    # Liters  - 6?
        self.min_volume = 1.5  # Liters - baloon starts slightly inflated.
        self.PC = 40           # Proportionality constant that relates pressure to cm-H2O
        self.P0 = 0            # Baseline/Minimum pressure.
        self.leak = leak

        self.temperature = 37  # keep track of these, as they are important output variable
        self.humidity = 90
        self.fio2 = 60

        # Dynamical parameters - these are the initial conditions
        self.current_flow     = 0  # in unit  liters/sec

        self.set_Qin          = 0  # set flow of a prop valve on inspiratory side      -- liters/second
        self.Qin              = 0  # exact flow of a prop valve on inspiratory side    -- liters/second

        self.set_Qout         = 0  # 0|max - setting of an solenoid on expiratory side -- liters/second
        self.Qout             = 0  # exact flow through the solenoid

        self.current_pressure = 0  # in unit  cm-H2O
        self.r_real = (3 * self.min_volume / (4 * np.pi)) ** (1 / 3)  # size of the lung
        self.current_volume = self.min_volume                         # in unit  liters

    def get_pressure(self):
        return self.current_pressure

    def get_volume(self):
        return self.current_volume

    def set_flow_in(self, Qin, dt):
        self.set_Qin = Qin

        Qin_clip = np.min([Qin, 2])                # Flows have to be positive, and reasonable. Nothing here is faster that 2 l/s
        Qin_clip = np.max([Qin_clip, 0])
        self.Qin     = Qin_clip                    # Assume the set-value is also the true value for prop

    def set_flow_out(self, Qout, dt):
        self.set_Qout = Qout

        Qout_clip = np.min([Qout, 2])                    # Flows have to be positive, and reasonable. Nothing here is faster that 2 l/s
        Qout_clip = np.max([Qout_clip, 0])
        difference_pressure = self.current_pressure - 0  # Convention: outside is "0"
        conductance = 0.05*Qout_clip                     # This should be in the range of ~1 liter/s for typical max pressure differences
        self.Qout = difference_pressure * conductance    # Target for flow out

    def update(self, dt):  # Performs an update of duration dt [seconds]

        if dt<1:                                        # This is the simulation, so not quite so important,
            self.current_flow = self.Qin - self.Qout     # But no update should take longer than that
            self.current_volume += self.current_flow * dt

            if self.leak:
                RC = 5  # pulled 5 sec out of my hat
                s = dt / (RC + dt)
                self.current_volume = self.current_volume + s * (self.min_volume - self.current_volume)

            # This is from the baloon equation, uses helper variable (the baloon radius)
            self.r_real = (3 * self.current_volume / (4 * np.pi)) ** (1 / 3)
            r0 = (3 * self.min_volume / (4 * np.pi)) ** (1 / 3)

            self.current_pressure = self.P0 + (self.PC / (r0 ** 2 * self.r_real)) * (1 - (r0 / self.r_real) ** 6)

            # Temperature, humidity and o2 fluctuations modelled as OUprocess
            self.temperature = self.OUupdate(self.temperature, dt=dt, mu=37, sigma=0.3, tau=1)
            self.fio2 = self.OUupdate(self.fio2, dt=dt, mu=60, sigma=5, tau=1)
            self.humidity = self.OUupdate(self.humidity, dt=dt, mu=90, sigma=5, tau=1)
            if self.humidity > 100:
                self.humidity = 100
        else:
            self._reset()
            print(self.current_pressure)


    def OUupdate(self, variable, dt, mu, sigma, tau):
        '''
        This is a simple function to produce an OU process.
        It is used as model for fluctuations in measurement variables.
        inputs:
        variable:   float     value at previous time step
        dt      :   timestep
        mu      :   mean
        sigma   :   noise amplitude
        tau     :   time scale
        returns:
        new_variable :  value of "variable" at next time step
        '''
        dt = max(dt, 0.05)  #Make sure this doesn't go haywire if anything hangs. Max 50ms
        sigma_bis = sigma * np.sqrt(2. / tau)
        sqrtdt = np.sqrt(dt)
        new_variable = variable + dt * (-(variable - mu) / tau) + sigma_bis * sqrtdt * np.random.randn()
        return new_variable

    def _reset(self):
        ''' resets the ballon to standard parameters. '''
        self.set_Qin          = 0
        self.Qin              = 0
        self.set_Qout         = 0
        self.Qout             = 0
        self.current_pressure = 0
        self.r_real = (3 * self.min_volume / (4 * np.pi)) ** (1 / 3)
        self.current_volume = self.min_volume

class ControlModuleSimulator(ControlModuleBase):
    # Implement ControlModuleBase functions
    def __init__(self):
        ControlModuleBase.__init__(self)
        self.Balloon = Balloon_Simulator(leak=False)          # This is the simulation
        self._sensor_to_COPY()

    def __SimulatedPropValve(self, x, dt):
        '''
        This simulates the action of a proportional valve.
        Flow-current-curve eye-balled from the datasheet of SMC PVQ31-5G-23-01N  
        https://www.ocpneumatics.com/content/pdfs/PVQ.pdf
        
        x:  Input current [mA]
        dt: Time since last setting in seconds [for the LP filter]
        '''
        flow_new = 1.0*(np.tanh(0.03*(x - 130)) + 1)
        if x>160:
            flow_new = 1.72  #Maximum, ~100 l/min
        if x<0:
            flow_new = 0
        return flow_new

    def __SimulatedSolenoid(self, x):
        '''
        Depending on x, set flow to a binary value.
        Here: flow is either 0 or 1l/sec
        '''
        if x > 0:
            return 1
        else:
            return 0

    def _sensor_to_COPY(self):
        # And the sensor measurements
        self._lock.acquire()
        self.COPY_sensor_values = SensorValues(vals={
            ValueName.PIP.name                  : self._DATA_PIP,
            ValueName.PEEP.name                 : self._DATA_PEEP,
            ValueName.FIO2.name                 : self.Balloon.fio2,
            ValueName.TEMP.name                 : self.Balloon.temperature,
            ValueName.HUMIDITY.name             : self.Balloon.humidity,
            ValueName.PRESSURE.name             : self.Balloon.current_pressure,
            ValueName.VTE.name                  : self._DATA_VTE,
            ValueName.BREATHS_PER_MINUTE.name   : self._DATA_BPM,
            ValueName.INSPIRATION_TIME_SEC.name : self._DATA_I_PHASE,
            'timestamp'                  : time.time(),
            'loop_counter'             : self._loop_counter,
            'breath_count': self._DATA_BREATH_COUNT
        })
        self._lock.release()

    def _start_mainloop(self):
        # start running, this should be run as a thread! 
        # Compare to initialization in Base Class!

        update_copies = self._NUMBER_CONTROLL_LOOPS_UNTIL_UPDATE

        while self._running.is_set():
            time.sleep(self._LOOP_UPDATE_TIME)
            self._loop_counter += 1
            now = time.time()
            dt = now - self._last_update                            # Time sincle last cycle of main-loop
            if dt > CONTROL[ValueName.BREATHS_PER_MINUTE].default / 4:                                                         # TODO: RAISE HARDWARE ALARM, no update should take that long
                print("Restarted cycle.")
                self._PID_reset()
                self.Balloon._reset()
                dt = self._LOOP_UPDATE_TIME

            self.Balloon.update(dt = dt)                            # Update the state of the balloon simulation
            self._DATA_PRESSURE = self.Balloon.get_pressure()       # Get a pressure measurement from balloon and tell controller             --- SENSOR 1

            self._PID_update(dt = dt)                               # Update the PID Controller

            x = self._get_control_signal_in()                       # Inspiratory side: get control signal for PropValve
            Qin = self.__SimulatedPropValve(x, dt = dt)             # And calculate the produced flow Qin

            y = self._get_control_signal_out()                      # Expiratory side: get control signal for Solenoid
            Qout = self.__SimulatedSolenoid(y)                      # Set expiratory flow rate, Qout

            self.Balloon.set_flow_in(Qin, dt = dt)                  # Set the flow rates for the Balloon simulator
            self.Balloon.set_flow_out(Qout, dt = dt)

            self._DATA_Qout = self.Balloon.Qout                     # Tell controller the expiratory flow rate, _DATA_Qout                    --- SENSOR 2
            self._DATA_Qin  = self.Balloon.Qin                      # Tell controller the expiratory flow rate, _DATA_Qin                     --- SENSOR 3
            self._last_update = now

            if update_copies == 0:
                self._controls_from_COPY()     # Update controls from possibly updated values as a chunk
                self._alarm_to_COPY()          # Copy current alarms and settings to COPY
                self._sensor_to_COPY()         # Copy sensor values to COPY
                update_copies = self._NUMBER_CONTROLL_LOOPS_UNTIL_UPDATE
            else:
                update_copies -= 1

        # # get final values on stop
        self._controls_from_COPY()  # Update controls from possibly updated values as a chunk
        self._alarm_to_COPY()  # Copy current alarms and settings to COPY
        self._sensor_to_COPY()  # Copy sensor values to COPY




def get_control_module(sim_mode=False):
    if sim_mode == True:
        return ControlModuleSimulator()
    else:
        return ControlModuleDevice()


import time
import sys
import threading
import pdb
import os

from PySide2 import QtWidgets, QtCore, QtGui


from vent.alarm import AlarmSeverity, Alarm
from vent.common import values
from vent.common.values import ValueName
from vent.common.message import ControlSetting
from vent.common.logging import init_logger
from vent import gui
from vent.gui import widgets, set_gui_instance, get_gui_instance, styles, PLOTS
from vent.gui.alarm_manager import AlarmManager




class Vent_Gui(QtWidgets.QMainWindow):

    gui_closing = QtCore.Signal()
    """
    :class:`PySide2.QtCore.Signal` emitted when the GUI is closing.
    """

    alarms_updated = QtCore.Signal(dict)
    """
    :class:`PySide2.QtCore.Signal` emitted whenever alarms are updated.
    
    Returns the result of ``self.coordinator.get_active_alarms``, so will emit an
    empty dict if there are no active alarms.
    """

    MONITOR = values.SENSOR
    """
    see :data:`.gui.defaults.SENSOR`
    """

    CONTROL = values.CONTROL
    """
    see :data:`.gui.defaults.CONTROL`
    """

    PLOTS = PLOTS
    """
    see :data:`.gui.defaults.PLOTS`
    """

    monitor_width = 2
    plot_width = 4
    control_width = 2
    total_width = monitor_width + plot_width + control_width
    """
    computed from ``monitor_width+plot_width+control_width``
    """

    status_height = 1
    main_height = 5
    total_height = status_height+main_height
    """
    computed from ``status_height+main_height``
    """

    def __init__(self, coordinator, update_period = 0.1):
        """
        The Main GUI window.

        Only one instance can be created at a time. Uses :func:`set_gui_instance` to
        store a reference to itself. after initialization, use `get_gui_instance` to
        retrieve a reference.

        .. todo:

            Jonny add a screenshot here when final version.


        Attributes:
            monitor (dict): Dictionary mapping :data:`.values.SENSOR` keys to :class:`.widgets.Monitor_Value` objects
            plots (dict): Dictionary mapping :data:`.gui.PLOT` keys to :class:`.widgets.Plot` objects
            controls (dict): Dictionary mapping :data:`.values.CONTROL` keys to :class:`.widgets.Control` objects
            coordinator (:class:`vent.coordinator.coordinator.CoordinatorBase`): Some coordinator object that we use to communicate with the controller
            control_module (:class:`vent.controller.control_module.ControlModuleBase`): Reference to the control module, retrieved from coordinator
            start_time (float): Start time as returned by :func:`time.time`
            update_period (float): The global delay between redraws of the GUI (seconds)
            alarm_manager (:class:`~.AlarmManager`)


        Arguments:
            update_period (float): The global delay between redraws of the GUI (seconds)
            test (bool): Whether the monitored values and plots should be fed sine waves for visual testing.


        """
        self.logger = init_logger(__name__)
        self.logger.info('gui init')

        if get_gui_instance() is not None and gui.limit_gui():
            raise Exception('Instance of gui already running!') # pragma: no cover
        else:
            set_gui_instance(self)

        super(Vent_Gui, self).__init__()

        self.alarm_manager = AlarmManager()
        self._alarm_state = AlarmSeverity.OFF

        self.monitor = {}
        self.plots = {}
        self.controls = {}

        self.coordinator = coordinator
        try:
            self.control_module = self.coordinator.control_module
        except AttributeError:
            self.control_module = None

        # start QTimer to update values
        self.timer = QtCore.QTimer()
        self.timer.setSingleShot(True)
        self.timer.timeout.connect(self.update_gui)
        # stop QTimer when program closing
        self.gui_closing.connect(self.timer.stop)

        # set update period (after timer is created!!)
        self._update_period = None
        self.update_period = update_period

        # initialize controls to starting values
        self.init_controls()

        self.init_ui()
        self.start_time = time.time()

        self.update_gui()

    @property
    def update_period(self):
        return self._update_period

    @update_period.setter
    def update_period(self, update_period):
        assert(isinstance(update_period, float) or isinstance(update_period, int))

        if update_period != self._update_period:
            # if the timer is active, stop it and restart with new period
            # if self.timer.isActive():
            self.timer.setInterval(update_period*1000)

            # store new value
            self._update_period = update_period

    def init_controls(self):
        """
        on startup, set controls in coordinator to ensure init state is synchronized
        """

        for control_name, control_params in self.CONTROL.items():
            self.set_value(control_params.default, control_name)

    def set_value(self, new_value, value_name=None):
        """
        Set a control value with the ``coordinator``

        Args:
            new_value (float): Som
        """
        # get sender ID
        if value_name is None:
            value_name = self.sender().objectName()

        elif not isinstance(value_name, str):
            # TODO: More explicitly check for enum
            value_name = value_name.name


        control_object = ControlSetting(name=getattr(ValueName, value_name),
                                        value=new_value,
                                        min_value = self.CONTROL[getattr(ValueName, value_name)]['safe_range'][0],
                                        max_value = self.CONTROL[getattr(ValueName, value_name)]['safe_range'][1],
                                        timestamp = time.time())
        self.coordinator.set_control(control_object)


    def update_gui(self):
        try:
            # get alarms
            #active_alarms = self.coordinator.get_active_alarms()
            #self.alarms_updated.emit(active_alarms)


            vals = self.coordinator.get_sensors()

            for plot_key, plot_obj in self.plots.items():
                if hasattr(vals, plot_key):
                    plot_obj.update_value((time.time(), getattr(vals, plot_key)))

            for monitor_key, monitor_obj in self.monitor.items():
                if hasattr(vals, monitor_key):
                    monitor_obj.update_value(getattr(vals, monitor_key))

        #
        finally:
            self.timer.start()




    def update_value(self, value_name, new_value):
        """
        Arguments:
            value_name (str): Name of key in :attr:`.Vent_Gui.monitor` and :attr:`.Vent_Gui.plots` to update
            new_value (int, float): New value to display/plot
        """
        if value_name in self.monitor.keys():
            self.monitor[value_name].update_value(new_value)
        elif value_name in self.plots.keys():
            self.plots[value_name].update_value(new_value)

    def init_ui(self):
        """
        Create the UI components for the ventilator screen
        """

        # basic initialization


        self.main_widget = QtWidgets.QWidget()
        self.main_widget.setContentsMargins(0,0,0,0)
        #
        self.setCentralWidget(self.main_widget)

        # layout
        #   - two rows (status, main)
        #   - three columns
        #       left:   monitor values
        #       center: plotted values
        #       right:  controls
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setContentsMargins(0,0,0,0)
        self.main_widget.setLayout(self.layout)

        # layout that includes the display and controls
        self.main_layout = QtWidgets.QHBoxLayout()
        self.main_layout.setContentsMargins(0,0,0,0)


        # call sub-create functions to make ui sections
        # top status bar
        self.init_ui_status_bar()
        # left monitored values
        self.init_ui_monitor()
        # plots
        self.init_ui_plots()
        # controls
        self.init_ui_controls()

        # add main ui area to layout
        self.layout.addLayout(self.main_layout, self.main_height)

        # connect signals and slots
        self.init_ui_signals()

        self.show()

    def init_ui_status_bar(self):
        ############
        # Status Bar
        status_box = QtWidgets.QGroupBox('System Status')
        status_box.setStyleSheet(styles.STATUS_BOX)
        status_layout = QtWidgets.QHBoxLayout()
        self.status_bar = widgets.Status_Bar()
        status_layout.addWidget(self.status_bar)
        status_layout.setContentsMargins(0,0,0,0)
        status_box.setLayout(status_layout)

        self.layout.addWidget(status_box, self.status_height)

    def init_ui_monitor(self):

        #########
        # display values
        # box that contains both the monitors and the plots
        self.monitor_box = QtWidgets.QGroupBox("Sensor Monitor")
        self.monitor_layout = QtWidgets.QHBoxLayout()
        self.monitor_layout.setContentsMargins(0, 0, 0, 0)
        self.monitor_box.setLayout(self.monitor_layout)

        # box that just displays the monitor widgets
        self.display_layout = QtWidgets.QVBoxLayout()
        self.display_layout.setContentsMargins(0,0,0,0)

        for display_key, display_params in self.MONITOR.items():
            self.monitor[display_key.name] = widgets.Monitor(display_params, enum_name=display_key)
            self.display_layout.addWidget(self.monitor[display_key.name])
            self.display_layout.addWidget(widgets.components.QHLine())

        self.display_layout.addStretch(10)

        self.monitor_layout.addLayout(self.display_layout, self.monitor_width)

        self.main_layout.addWidget(self.monitor_box, self.plot_width + self.monitor_width)


    def init_ui_plots(self):
        ###########
        # plots
        self.plot_layout = QtWidgets.QVBoxLayout()
        self.plot_layout.setContentsMargins(0, 0, 0, 0)

        # button to set plot history
        button_box = QtWidgets.QGroupBox("Plot History")
        # button_group = QtWidgets.QButtonGroup()
        # button_group.exclusive()
        times = (("5s", 5),
                 ("10s", 10),
                 ("30s", 30),
                 ("1m", 60),
                 ("5m", 60 * 5),
                 ("15m", 60 * 15),
                 ("60m", 60 * 60))

        self.time_buttons = {}
        button_layout = QtWidgets.QHBoxLayout()
        button_layout.addStretch()

        for a_time in times:
            self.time_buttons[a_time[0]] = QtWidgets.QRadioButton(a_time[0])
            self.time_buttons[a_time[0]].setObjectName(str(a_time[1]))
            self.time_buttons[a_time[0]].clicked.connect(self.set_plot_duration)
            button_layout.addWidget(self.time_buttons[a_time[0]])

        button_box.setLayout(button_layout)
        self.plot_layout.addWidget(button_box)

        # the plot widgets themselves
        for plot_key, plot_params in self.PLOTS.items():
            self.plots[plot_key.name] = widgets.Plot(**plot_params)
            self.plot_layout.addWidget(self.plots[plot_key.name])

        # self.main_layout.addLayout(self.plot_layout,5)
        self.monitor_layout.addLayout(self.plot_layout, self.plot_width)

        ######################
        # set the default view as 30s
        self.time_buttons[times[2][0]].click()

    def init_ui_controls(self):
        ####################
        # Controls
        self.controls_box = QtWidgets.QGroupBox("Ventilator Controls")
        # set name so it catches the stylesheet
        self.controls_box.setObjectName('CONTROLBOX')
        # controls_box.setStyleSheet(styles.CONTROL_BOX)

        self.controls_box.setContentsMargins(0, 0, 0, 0)

        self.controls_layout = QtWidgets.QVBoxLayout()
        self.controls_layout.setContentsMargins(0, 0, 0, 0)
        for control_name, control_params in self.CONTROL.items():
            self.controls[control_name.name] = widgets.Control(control_params)
            self.controls[control_name.name].setObjectName(control_name.name)
            self.controls[control_name.name].value_changed.connect(self.set_value)
            self.controls_layout.addWidget(self.controls[control_name.name])
            self.controls_layout.addWidget(widgets.components.QHLine(color=styles.DIVIDER_COLOR_DARK))

        self.controls_layout.addStretch(10)
        self.controls_box.setLayout(self.controls_layout)

        self.main_layout.addWidget(self.controls_box, self.control_width)

    def init_ui_signals(self):
        """
        Connect Qt signals and slots
        """

        # hook up monitors and plots
        for value in ValueName:
            if value.name in self.plots.keys():
                self.monitor[value.name].limits_changed.connect(
                    self.plots[value.name].set_safe_limits)
                self.plots[value.name].limits_changed.connect(
                    self.monitor[value.name].update_limits)

        # connect monitors to alarm_manager
        for monitor in self.monitor.values():
            monitor.alarm.connect(self.alarm_manager.monitor_alarm)

        # connect alarms to alarm manager, and then back to us
        self.alarms_updated.connect(self.alarm_manager.update_alarms)
        self.alarm_manager.new_alarm.connect(self.handle_alarm)
        # FIXME: THis should be handled by alarm manager, that's what it's there for!
        #self.status_bar.status_message.level_changed.connect(self.alarm_state_changed)
        self.status_bar.status_message.message_cleared.connect(self.handle_cleared_alarm)

        # connect start button to coordinator start
        self.status_bar.start_button.clicked.connect(self.coordinator.start)

    @QtCore.Slot(Alarm)
    def handle_alarm(self, alarm):
        self.status_bar.status_message.update_message(alarm)
        try:
            self.monitor[alarm.alarm_name.name].alarm_state = True
        except:
            # FIXME: will be fixed when values are displayed next to controls
            pass
        if alarm.severity.value > self.alarm_state.value:
            self.alarm_state = alarm.severity

    @QtCore.Slot(Alarm)
    def handle_cleared_alarm(self, alarm):
        try:
            self.monitor[alarm.alarm_name.name].alarm_state = False
        except:
            # FIXME: will be fixed when values are displayed next to controls
            pass


    @property
    def alarm_state(self):
        return self._alarm_state

    @alarm_state.setter
    def alarm_state(self, state):
        if state == AlarmSeverity.HIGH:
            pass

    @QtCore.Slot(AlarmSeverity)
    def alarm_state_changed(self, state):
        self.alarm_state = state

    def set_plot_duration(self, dur):
        dur = int(self.sender().objectName())

        for plot in self.plots.values():
            plot.set_duration(dur)

    def closeEvent(self, event):
        """
        Emit :attr:`.gui_closing` and close!
        """
        #globals()['_GUI_INSTANCE'] = None
        set_gui_instance(None)
        self.gui_closing.emit()

        if self.coordinator:
            try:
                self.coordinator.stop()
            except:
                raise Warning('had a thread object, but the thread object couldnt be stopped')

        event.accept()

    def start(self):
        """
        Click the :meth:`~.gui.widgets.status_bar.Status_Bar.start` button

        Returns:

        """
        self.status_bar.start_button.click()


def launch_gui(coordinator):

    # just for testing, should be run from main
    app = QtWidgets.QApplication(sys.argv)
    app.setStyle('Fusion')
    app.setStyleSheet(styles.DARK_THEME)
    app = styles.set_dark_palette(app)
    gui = Vent_Gui(coordinator)

    return app, gui


""" Logging functionality

"""
import traceback
import os
import logging
from logging import handlers

# some global stack param
MAX_STACK_DEPTH = 20

from vent import prefs


def init_logger(module_name: str,
                log_level: int = logging.DEBUG,
                file_handler: bool = True) -> logging.Logger:
    """
    Initialize a

    Args:
        module_name (str): module name used to generate filename and name logger
        log_level (int): one of :var:`logging.DEBUG`, :var:`logging.INFO`, :var:`logging.WARNING`, or :var:`logging.ERROR`
        file_handler (bool, str): if ``True``, (default), log in ``<logdir>/module_name.log`` .
            if ``False``, don't log to disk.

    Returns:
        :class:`logging.Logger` : Logger 4 u 2 use
    """
    logger = logging.getLogger(module_name)

    # set log level
    assert log_level in (logging.DEBUG,
                         logging.INFO,
                         logging.WARNING,
                         logging.ERROR)
    logger.setLevel(log_level)

    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    # I assume this is to stop printing to stderr? why does it get a formatter then? -jls 2020-05-25
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    # handler to log to disk
    # max = 8 file x 16 MB = 128 MB
    if file_handler:
        log_filename = os.path.join(prefs.get_pref('LOG_DIR'),
                                    module_name + '.log')
        fh = logging.handlers.RotatingFileHandler(
            log_filename,
            mode = 'a',
            maxBytes=16 * 2 ** 20,
            backupCount=7
        )
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    return logger


def log_exception(e, tb):
    """  # TODO: Stub exception logger. Prints exception type & traceback

    Args:
        e: Exception to log
        tb: TraceBack associated with Exception e

    """
    print("Caught the following exception:", e, " but I don't know what to do with it.")
    # print(traceback.print_tb(tb, limit=MAX_STACK_DEPTH))
    raise



from vent.gui.widgets.control import Control
from vent.gui.widgets.monitor import Monitor
from vent.gui.widgets.plot import Plot
from vent.gui.widgets.status_bar import Status_Bar
from vent.gui.widgets import components


import time
from collections import deque

import numpy as np
from PySide2 import QtCore
import PySide2 # import so pyqtgraph recognizes as what we're using
import pyqtgraph as pg


from vent.gui import styles
from vent.gui import mono_font
from vent.gui import get_gui_instance

PLOT_TIMER = None
"""
A :class:`~PySide2.QtCore.QTimer` that updates :class:`.TimedPlotCurveItem`s
"""

PLOT_FREQ = 5
"""
Update frequency of :class:`.Plot` s in Hz
"""


class Plot(pg.PlotWidget):

    limits_changed = QtCore.Signal(tuple)

    def __init__(self, name, buffer_size = 4092, plot_duration = 5, abs_range = None, safe_range = None, color=None, units='', **kwargs):
        #super(Plot, self).__init__(axisItems={'bottom':TimeAxis(orientation='bottom')})
        # construct title html string
        titlestr = "<h1 style=\"{title_style}\">{title_text} ({units})</h1>".format(title_style=styles.TITLE_STYLE,
                                                                      title_text=name,
                                                                        units =units)


        super(Plot, self).__init__(background=styles.BACKGROUND_COLOR,
                                   title=titlestr)
        self.timestamps = deque(maxlen=buffer_size)
        self.history = deque(maxlen=buffer_size)
        # TODO: Make @property to update buffer_size, preserving history
        self.plot_duration = plot_duration

        self.units = units


        self._start_time = time.time()
        self._last_time = time.time()
        self._last_relative_time = 0

        self.abs_range = None
        if abs_range:
            self.abs_range = abs_range
            self.setYRange(self.abs_range[0], self.abs_range[1])

        self.safe_range = (0,0)
        if safe_range:
            self.safe_range = safe_range


        self.setXRange(0, plot_duration)

        # split plot curve into two so that the endpoint doesn't get connected to the start point
        self.early_curve = self.plot(width=3)
        self.late_curve = self.plot(width=3)
        self.time_marker = self.plot()

        self.min_safe = pg.InfiniteLine(movable=True, angle=0, pos=self.safe_range[0])
        self.max_safe = pg.InfiniteLine(movable=True, angle=0, pos=self.safe_range[1])
        self.min_safe.sigPositionChanged.connect(self._safe_limits_changed)
        self.max_safe.sigPositionChanged.connect(self._safe_limits_changed)

        self.addItem(self.min_safe)
        self.addItem(self.max_safe)

        if color:
            self.early_curve.setPen(color=color, width=3)
            self.late_curve.setPen(color=color, width=3)


    def set_duration(self, dur):
        self.plot_duration = int(round(dur))
        self.setXRange(0, self.plot_duration)


    def update_value(self, new_value):
        """
        new_value: (timestamp from time.time(), value)
        """
        try:
            this_time = time.time()
            #time_diff = this_time-self._last_time
            limits = self.getPlotItem().viewRange()
            current_relative_time = (this_time-self._start_time) % self.plot_duration
            self.time_marker.setData([current_relative_time, current_relative_time],
                                     [limits[1][0], limits[1][1]])

            self.timestamps.append(new_value[0])
            self.history.append(new_value[1])

            # filter values based on timestamps
            ts_array = np.array(self.timestamps)
            end_ind = len(self.history)
            start_ind = np.where(ts_array > (this_time - self.plot_duration))[0][0]

            # subtract start time and take modulus of duration to get wrapped timestamps
            plot_timestamps = np.mod(ts_array[start_ind:end_ind]-self._start_time, self.plot_duration)
            plot_values = np.array([self.history[i] for i in range(start_ind, end_ind)])

            # find the point where the time resets
            try:
                reset_ind = np.where(np.diff(plot_timestamps)<0)[0][0]

                # plot early and late
                self.early_curve.setData(plot_timestamps[0:reset_ind+1],plot_values[0:reset_ind+1] )
                self.late_curve.setData(plot_timestamps[reset_ind+1:], plot_values[reset_ind+1:])

            except IndexError:
                self.early_curve.setData(plot_timestamps, plot_values)
                self.late_curve.clear()
        except:
            # FIXME: Log this lol
            print('error plotting value: {}, timestamp: {}'.format(new_value[1], new_value[0]))

        #self._last_time = this_time

    def _safe_limits_changed(self, val):
        # ignore input val, just emit the current value of the lines
        self.limits_changed.emit((self.min_safe.value(),
                                       self.max_safe.value()))

    @QtCore.Slot(tuple)
    def set_safe_limits(self, limits):
        self.max_safe.setPos(limits[1])
        self.min_safe.setPos(limits[0])
    #
    # def plot(self, *args, **kargs):
    #     """
    #     Override method :meth:`pyqtgraph.graphicsItems.PlotItem.plot` to return :class:`TimedPlotDataItem`
    #
    #     Add and return a new plot.
    #     See :func:`PlotDataItem.__init__ <pyqtgraph.PlotDataItem.__init__>` for data arguments
    #
    #     Extra allowed arguments are:
    #         clear    - clear all plots before displaying new data
    #         params   - meta-parameters to associate with this data
    #     """
    #     clear = kargs.get('clear', False)
    #     params = kargs.get('params', None)
    #
    #     if clear:
    #         self.clear()
    #
    #     item = TimedPlotDataItem(*args, **kargs)
    #
    #     if params is None:
    #         params = {}
    #     self.addItem(item, params=params)
    #
    #     return item
#
# class TimedPlotDataItem(pg.PlotDataItem):
#     """
#     Subclass of :class:`pyqtgraph.PlotDataItem` that uses :class:`.TimedPlotCurveItem`
#
#     """
#
#     def __init__(self, *args, **kwargs):
#         super(TimedPlotDataItem, self).__init__(*args, **kwargs)
#
#         # replace self.curve with timed version
#         # mimic __init__ from parent
#         # https://github.com/pyqtgraph/pyqtgraph/blob/a2053b13d0234e210561a73fa044625fd972e910/pyqtgraph/graphicsItems/PlotDataItem.py#L37
#         self.curve = TimedPlotCurveItem()
#         self.curve.setParentItem(self)
#         self.curve.sigClicked.connect(self.curveClicked)
#
#         self.setData(*args, **kwargs)
#
# class TimedPlotCurveItem(pg.PlotCurveItem):
#     """
#     Subclass :class:`pyqtgraph.PlotCurveItem` to update with a fixed frequency instead of whenever new data pushed
#
#     Either creates or gets :data:`PLOT_TIMER` and connects it to :meth:`TimedPlotCurveItem.update`
#     """
#
#
#     def __init__(self, *args, **kwargs):
#         super(TimedPlotCurveItem, self).__init__(*args, **kwargs)
#
#         if globals()['PLOT_TIMER'] is None:
#             globals()['PLOT_TIMER'] = QtCore.QTimer()
#
#             get_gui_instance().gui_closing.connect(globals()['PLOT_TIMER'].stop)
#
#         self.timer = globals()['PLOT_TIMER']
#         # stop timer, add ourselves, restart
#         self.timer.timeout.connect(self.update)
#
#         try:
#             self.fps = float(globals()['PLOT_FREQ'])
#         except (ValueError, KeyError):
#             self.fps = 15.
#
#         self.timer.start(1./self.fps)
#
#
#
#     def __del__(self):
#         """
#         On object deletion, remove from :data:`.PLOT_TIMER` 's signals
#         """
#         #self.timer.disconnect(self.update)
#         #super(TimedPlotCurveItem, self).__del__()
#
#
#     def updateData(self, *args, **kargs):
#         """
#         Override :meth:`pyqtgraph.PlotCurveItem.updateData` to omit call to :meth:`.update`
#         which is done by :data:`.PLOT_TIMER`
#
#         """
#         profiler = pg.debug.Profiler()
#
#         if 'compositionMode' in kargs:
#             self.setCompositionMode(kargs['compositionMode'])
#
#         if len(args) == 1:
#             kargs['y'] = args[0]
#         elif len(args) == 2:
#             kargs['x'] = args[0]
#             kargs['y'] = args[1]
#
#         if 'y' not in kargs or kargs['y'] is None:
#             kargs['y'] = np.array([])
#         if 'x' not in kargs or kargs['x'] is None:
#             kargs['x'] = np.arange(len(kargs['y']))
#
#         for k in ['x', 'y']:
#             data = kargs[k]
#             if isinstance(data, list):
#                 data = np.array(data)
#                 kargs[k] = data
#             if not isinstance(data, np.ndarray) or data.ndim > 1:
#                 raise Exception("Plot data must be 1D ndarray.")
#             if data.dtype.kind == 'c':
#                 raise Exception("Can not plot complex data types.")
#
#         profiler("data checks")
#
#         # self.setCacheMode(QtGui.QGraphicsItem.NoCache)  ## Disabling and re-enabling the cache works around a bug in Qt 4.6 causing the cached results to display incorrectly
#         ##    Test this bug with test_PlotWidget and zoom in on the animated plot
#         self.yData = kargs['y'].view(np.ndarray)
#         self.xData = kargs['x'].view(np.ndarray)
#
#         self.invalidateBounds()
#         self.prepareGeometryChange()
#         self.informViewBoundsChanged()
#
#         profiler('copy')
#
#         if 'stepMode' in kargs:
#             self.opts['stepMode'] = kargs['stepMode']
#
#         if self.opts['stepMode'] is True:
#             if len(self.xData) != len(self.yData) + 1:  ## allow difference of 1 for step mode plots
#                 raise Exception("len(X) must be len(Y)+1 since stepMode=True (got %s and %s)" % (
#                 self.xData.shape, self.yData.shape))
#         else:
#             if self.xData.shape != self.yData.shape:  ## allow difference of 1 for step mode plots
#                 raise Exception(
#                     "X and Y arrays must be the same shape--got %s and %s." % (self.xData.shape, self.yData.shape))
#
#         self.path = None
#         self.fillPath = None
#         self._mouseShape = None
#         # self.xDisp = self.yDisp = None
#
#         if 'name' in kargs:
#             self.opts['name'] = kargs['name']
#         if 'connect' in kargs:
#             self.opts['connect'] = kargs['connect']
#         if 'pen' in kargs:
#             self.setPen(kargs['pen'])
#         if 'shadowPen' in kargs:
#             self.setShadowPen(kargs['shadowPen'])
#         if 'fillLevel' in kargs:
#             self.setFillLevel(kargs['fillLevel'])
#         if 'fillOutline' in kargs:
#             self.opts['fillOutline'] = kargs['fillOutline']
#         if 'brush' in kargs:
#             self.setBrush(kargs['brush'])
#         if 'antialias' in kargs:
#             self.opts['antialias'] = kargs['antialias']
#
#         profiler('set')
#         #self.update()
#         profiler('update')
#         #self.sigPlotChanged.emit(self)
#         profiler('emit')




import copy

import numpy as np
from PySide2 import QtWidgets, QtCore, QtGui

from vent.gui import styles, mono_font

class DoubleSlider(QtWidgets.QSlider):
    """
    Slider capable of representing floats

    Ripped off from
    and https://stackoverflow.com/a/50300848 ,

    Thank you!!!
    """

    # override the default valueChanged signal
    doubleValueChanged = QtCore.Signal(float)

    def __init__(self, decimals=1, *args, **kargs):
        super(DoubleSlider, self).__init__(*args, **kargs)
        self._multi = 10 ** decimals
        self.decimals = decimals

        self.valueChanged.connect(self.emitDoubleValueChanged)

    def setDecimals(self, decimals):
        self._multi = 10 ** decimals

    def emitDoubleValueChanged(self):
        self.doubleValueChanged.emit(self.value())

    def value(self):
        return float(super(DoubleSlider, self).value()) / self._multi

    def setMinimum(self, value):
        return super(DoubleSlider, self).setMinimum(round(value * self._multi))

    def setMaximum(self, value):
        return super(DoubleSlider, self).setMaximum(round(value * self._multi))

    def minimum(self):
        return super(DoubleSlider, self).minimum() / self._multi

    def _minimum(self):
        return super(DoubleSlider, self).minimum()

    def maximum(self):
        return super(DoubleSlider, self).maximum() / self._multi

    def _maximum(self):
        return super(DoubleSlider, self).maximum()

    def setSingleStep(self, value):
        return super(DoubleSlider, self).setSingleStep(round(value * self._multi))

    def singleStep(self):
        return float(super(DoubleSlider, self).singleStep()) / self._multi

    def _singleStep(self):
        return super(DoubleSlider, self).singleStep()

    def setValue(self, value):

        super(DoubleSlider, self).setValue(int(round(value * self._multi)))


class RangeSlider(DoubleSlider):

    valueChanged = QtCore.Signal(tuple)
    """
    (tuple): (low, high) set range of floats
    """

    def __init__(self, abs_range, safe_range, decimals=1, *args, **kwargs):
        """
        Slider with two handles that sets a range

        Args:
            abs_range (tuple): absolute range of slider
            safe_range (tuple): default set values for handles of slider
            decimals (int): number of decimals of precision
            *args:
            **kwargs:
        """
        super(RangeSlider, self).__init__(decimals=decimals, *args, **kwargs)
        self.setStyleSheet(styles.RANGE_SLIDER)

        self.decimals = decimals
        self.setSingleStep(10 ** -self.decimals)

        # abs range is the min/max theoretically allowable values
        self.abs_range = abs_range
        self.setMinimum(abs_range[0])
        self.setMaximum(abs_range[1])

        # safe range are the levels outside of which we trigger an alarm
        # initialize high and low to full range first so they have acceptable values
        self._low = int(round(abs_range[0] * self._multi))
        self._high = int(round(abs_range[1] * self._multi))
        # then set the initial value
        self.safe_range = safe_range
        self.setValue(self.safe_range)

        self.pressed_control = QtWidgets.QStyle.SC_None
        self.hover_control = QtWidgets.QStyle.SC_None
        self.click_offset = 0

        # 0 for the low, 1 for the high, -1 for both
        self.active_slider = 0

        # cache tick levels and labels so they're not redrawn constantly
        self.levels = None
        self.level_labels = None

        if self.orientation() == QtCore.Qt.Orientation.Horizontal:
            self.setMinimumHeight(styles.SLIDER_HEIGHT)
        else:
            self.setMinimumWidth(styles.SLIDER_WIDTH)


    @property
    def low(self):
        return self._low / self._multi

    @low.setter
    def low(self, low):
        old_low = self._low

        low = int(round(low * self._multi))

        # if the low value is above the minimum...
        if low >= self._minimum():
            # if the low value is higher than the high value,
            # and there is at least one step of headroom,
            # set the high value as well
            if low > self._high:
                if low <= self._maximum() - self._singleStep():
                    self._high = low + self._singleStep()
                else:
                    return
            self._low = low

        elif low < self._minimum():
            self._low = self._minimum()

        if old_low != self._low:
            self.valueChanged.emit((self.low, self.high))
        self.update()

    @property
    def high(self):
        return self._high / self._multi

    @high.setter
    def high(self, high):
        old_high = self._high

        high = int(round(high * self._multi))

        # if the low value is above the minimum...
        if high <= self._maximum():
            # if the high value is lower than the low value,
            # and there is at least one step of headroom,
            # set the high value as well
            if high < self._low:
                if high >= self._minimum() + self._singleStep():
                    self._low = high - self._singleStep()
                else:
                    return
            self._high = high

        elif high > self._maximum():
            self._high = self._maximum()

        if old_high != self._high:
            self.valueChanged.emit((self.low, self.high))

        self.update()

    # make methods just so the can accept signals
    def setLow(self, low):
        self.low = low

    def setHigh(self, high):
        self.high = high

    def setValue(self, value):
        """
        Args:
            value (tuple): (low, high) to set
        """

        # if our signals were not blocked before now,
        # we want to turn them on again afterwards

        old_state = self.signalsBlocked()

        # block signals so we don't double-emit the change
        self.blockSignals(True)
        self.low = value[0]
        self.high = value[1]
        self.blockSignals(old_state)

        self.valueChanged.emit((self.low, self.high))

    def value(self):
        return (self.low, self.high)

    def generate_labels(self):
        """
        Generate the text labels for the slider.

        Called on init and on resizeEvent
        """


        self.levels = np.linspace(self.minimum(), self.maximum(), 5)
        self.level_labels = []

        painter = QtGui.QPainter(self)
        painter.setFont(mono_font())

        style = self.style()
        opt = QtWidgets.QStyleOptionSlider()
        self.initStyleOption(opt)
        length = style.pixelMetric(QtWidgets.QStyle.PM_SliderLength, opt, self)
        available = style.pixelMetric(QtWidgets.QStyle.PM_SliderSpaceAvailable, opt, self)

        for v in self.levels:
            label_str = str(int(round(v)))
            # label_str = "{0:d}".format(v)
            rect = painter.drawText(QtCore.QRect(), QtCore.Qt.TextDontPrint, label_str)

            if self.orientation() == QtCore.Qt.Horizontal:

                x_loc = QtWidgets.QStyle.sliderPositionFromValue(self.minimum(),
                                                                 self.maximum(), v, available, opt.upsideDown)

                left = x_loc + length // 2 - rect.width() // 2
                # there is a 3 px offset that I can't attribute to any metric
                # left = (self.width())-(rect.width())-10
                bottom = (self.height() / 2) + rect.height() //2 + 3

            else:

                y_loc= QtWidgets.QStyle.sliderPositionFromValue(self.minimum(),
                        self.maximum(), v, available, opt.upsideDown)

                bottom=y_loc+length//2+rect.height()//2-3
                # there is a 3 px offset that I can't attribute to any metric
                #left = (self.width())-(rect.width())-10
                left = (self.width()/2)-(styles.INDICATOR_WIDTH/2)-rect.width()-3

            pos=QtCore.QPoint(left, bottom)
            self.level_labels.append((pos, QtGui.QStaticText(label_str)))

            #painter.drawText(pos, label_str)

        self.setTickInterval(self.levels[1]-self.levels[0])
        self.redraw_labels = False


    def paintEvent(self, event):
        # based on http://qt.gitorious.org/qt/qt/blobs/master/src/gui/widgets/qslider.cpp

        painter = QtGui.QPainter(self)
        #style = QtWidgets.QApplication.style()
        style = self.style()

        for i, value in enumerate([self._high, self._low]):
            opt = QtWidgets.QStyleOptionSlider()
            self.initStyleOption(opt)
            # pdb.set_trace()
            # Only draw the groove for the first slider so it doesn't get drawn
            # on top of the existing ones every time
            if i == 0:
                opt.subControls = style.SC_SliderGroove | style.SC_SliderHandle
            else:
                # opt.subControls = QtWidgets.QStyle.SC_SliderHandle
                opt.subControls = style.SC_SliderHandle

            #
            # if self.tickPosition() != self.NoTicks:
            #     opt.subControls |= QtWidgets.QStyle.SC_SliderTickmarks

            if self.pressed_control:
                opt.activeSubControls = self.pressed_control
                opt.state |= QtWidgets.QStyle.State_Sunken
            else:
                opt.activeSubControls = self.hover_control

            # opt.rect.setX(-self.width()/2)
            opt.sliderPosition = value
            opt.sliderValue = value
            style.drawComplexControl(QtWidgets.QStyle.CC_Slider, opt, painter, self)

        if not self.level_labels or self.redraw_labels:
            painter.end()
            self.generate_labels()
            painter = QtGui.QPainter(self)

        painter.setFont(mono_font())

        for label in self.level_labels:
            painter.drawStaticText(label[0], label[1])

        #self.setTickInterval(self.levels[1]-self.levels[0])

    def get_handle_rect(self, which=0):
        """

        Args:
            which (int): 0 = low, 1 = high

        Returns:
            :class:`QtCore.QRect` of handle
        """

        painter = QtGui.QPainter(self)
        #style = QtWidgets.QApplication.style()
        style = self.style()

        if which == 0 or which is False:
            value = self._low
        elif which == 1 or which is True:
            value = self._high
        else:
            raise ValueError(f'needs to be 0 for low handle or 1 for high handle, got {which}')


        opt = QtWidgets.QStyleOptionSlider()
        self.initStyleOption(opt)
        opt.subControls = style.SC_SliderHandle
        opt.sliderPosition = value
        opt.sliderValue = value

        return style.subControlRect(QtWidgets.QStyle.CC_Slider, opt, QtWidgets.QStyle.SC_SliderHandle, self)






    def mousePressEvent(self, event):
        event.accept()

        style = QtWidgets.QApplication.style()
        button = event.button()

        # In a normal slider control, when the user clicks on a point in the
        # slider's total range, but not on the slider part of the control the
        # control would jump the slider value to where the user clicked.
        # For this control, clicks which are not direct hits will slide both
        # slider parts

        if button:
            opt = QtWidgets.QStyleOptionSlider()
            self.initStyleOption(opt)

            self.active_slider = -1

            for i, value in enumerate([self._low, self._high]):
                opt.sliderPosition = value
                hit = style.hitTestComplexControl(style.CC_Slider, opt, event.pos(), self)
                if hit == style.SC_SliderHandle:
                    self.active_slider = i
                    self.pressed_control = hit

                    self.triggerAction(self.SliderMove)
                    self.setRepeatAction(self.SliderNoAction)
                    self.setSliderDown(True)
                    break

            if self.active_slider < 0:
                self.pressed_control = QtWidgets.QStyle.SC_SliderHandle
                self.click_offset = self.__pixelPosToRangeValue(self.__pick(event.pos()))
                self.triggerAction(self.SliderMove)
                self.setRepeatAction(self.SliderNoAction)
        else:
            event.ignore()

    def mouseMoveEvent(self, event):
        if self.pressed_control != QtWidgets.QStyle.SC_SliderHandle:
            event.ignore()
            return

        event.accept()

        # get old values
        old_low = copy.copy(self.low)
        old_high = copy.copy(self.high)

        new_pos = self.__pixelPosToRangeValue(self.__pick(event.pos()))
        opt = QtWidgets.QStyleOptionSlider()
        self.initStyleOption(opt)
        if self.active_slider < 0:
            offset = new_pos - self.click_offset
            self.high += offset
            self.low += offset
            if self.low < self.minimum():
                diff = self.minimum() - self.low
                self.low += diff
                self.high += diff
            if self.high > self.maximum():
                diff = self.maximum() - self.high
                self.low += diff
                self.high += diff
        elif self.active_slider == 0:
            if new_pos >= self.high:
                #new_pos = self._high - 1
                new_pos = self.low
            self.low = new_pos
        else:
            if new_pos <= self.low:
                #new_pos = self._low + 1
                new_pos = self.high
            self.high = new_pos
        self.click_offset = new_pos
        self.update()
        self.emit(QtCore.SIGNAL('sliderMoved(int)'), new_pos)

        # emit valuechanged signal
        if (old_low != self.low) or (old_high != self.high):
            self.valueChanged.emit((self.low, self.high))

    def __pick(self, pt):
        if self.orientation() == QtCore.Qt.Horizontal:
            return pt.x()
        else:
            return pt.y()

    def __pixelPosToRangeValue(self, pos):
        opt = QtWidgets.QStyleOptionSlider()
        self.initStyleOption(opt)
        style = QtWidgets.QApplication.style()

        gr = style.subControlRect(style.CC_Slider, opt, style.SC_SliderGroove, self)
        sr = style.subControlRect(style.CC_Slider, opt, style.SC_SliderHandle, self)

        if self.orientation() == QtCore.Qt.Horizontal:
            slider_length = sr.width()
            slider_min = gr.x()
            slider_max = gr.right() - slider_length + 1
        else:
            slider_length = sr.height()
            slider_min = gr.y()
            slider_max = gr.bottom() - slider_length + 1
        return style.sliderValueFromPosition(self.minimum()*self._multi, self.maximum()*self._multi, pos - slider_min, slider_max - \
            slider_min, opt.upsideDown)/self._multi

    def resizeEvent(self, event):
        self.redraw_labels = True



class KeyPressHandler(QtCore.QObject):
    """Custom key press handler
    https://gist.github.com/mfessenden/baa2b87b8addb0b60e54a11c1da48046"""
    escapePressed = QtCore.Signal(bool)
    returnPressed = QtCore.Signal(bool)

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.KeyPress:
            event_key = event.key()
            if event_key == QtCore.Qt.Key_Escape:
                self.escapePressed.emit(True)
                return True
            if event_key == QtCore.Qt.Key_Return or event_key == QtCore.Qt.Key_Enter:
                self.returnPressed.emit(True)
                return True

        return QtCore.QObject.eventFilter(self, obj, event)


class EditableLabel(QtWidgets.QWidget):
    """Editable label
    https://gist.github.com/mfessenden/baa2b87b8addb0b60e54a11c1da48046"""
    textChanged = QtCore.Signal(str)

    def __init__(self, parent=None, **kwargs):
        super(EditableLabel, self).__init__(parent=parent, **kwargs)

        self.is_editable = kwargs.get("editable", True)
        self.keyPressHandler = KeyPressHandler(self)

        # self.setStyleSheet(styles.CONTROL_VALUE)

        self.mainLayout = QtWidgets.QHBoxLayout(self)
        self.mainLayout.setContentsMargins(0, 0, 0, 0)
        self.mainLayout.setObjectName("mainLayout")

        self.label = QtWidgets.QLabel(self)
        self.label.setObjectName("label")
        # self.label.setStyleSheet(styles.CONTROL_VALUE)
        # self.label.setMinimumHeight(styles.VALUE_SIZE)


        self.mainLayout.addWidget(self.label)
        self.lineEdit = QtWidgets.QLineEdit(self)
        # self.lineEdit.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
        #                             QtWidgets.QSizePolicy.Expanding)
        # self.lineEdit.setMinimumHeight(styles.VALUE_SIZE)
        self.lineEdit.setObjectName("lineEdit")
        self.lineEdit.setValidator(QtGui.QDoubleValidator())
        self.mainLayout.addWidget(self.lineEdit)
        # hide the line edit initially
        self.lineEdit.setHidden(True)


        # setup signals
        self.create_signals()

    def create_signals(self):
        self.lineEdit.installEventFilter(self.keyPressHandler)
        self.label.mousePressEvent = self.labelPressedEvent

        # give the lineEdit both a `returnPressed` and `escapedPressed` action
        self.keyPressHandler.escapePressed.connect(self.escapePressedAction)
        self.keyPressHandler.returnPressed.connect(self.returnPressedAction)

    def text(self):
        """Standard QLabel text getter"""
        return self.label.text()

    def setText(self, text):
        """Standard QLabel text setter"""
        self.label.blockSignals(True)
        self.label.setText(text)
        self.label.blockSignals(False)

    def labelPressedEvent(self, event):
        """Set editable if the left mouse button is clicked"""
        if event.button() == QtCore.Qt.MouseButton.LeftButton:
            self.setLabelEditableAction()

    def setLabelEditableAction(self):
        """Action to make the widget editable"""
        if not self.is_editable:
            return

        self.label.setHidden(True)
        self.label.blockSignals(True)
        self.lineEdit.setHidden(False)
        self.lineEdit.setText(self.label.text())
        self.lineEdit.blockSignals(False)
        self.lineEdit.setFocus(QtCore.Qt.MouseFocusReason)
        self.lineEdit.selectAll()

    def labelUpdatedAction(self):
        """Indicates the widget text has been updated"""
        text_to_update = self.lineEdit.text()

        if text_to_update != self.label.text():
            self.label.setText(text_to_update)
            self.textChanged.emit(text_to_update)

        self.label.setHidden(False)
        self.lineEdit.setHidden(True)
        self.lineEdit.blockSignals(True)
        self.label.blockSignals(False)

    def returnPressedAction(self):
        """Return/enter event handler"""
        self.labelUpdatedAction()

    def escapePressedAction(self):
        """Escape event handler"""
        self.label.setHidden(False)
        self.lineEdit.setHidden(True)
        self.lineEdit.blockSignals(True)
        self.label.blockSignals(False)


class QHLine(QtWidgets.QFrame):
    """
    with respct to https://stackoverflow.com/a/51057516
    """
    def __init__(self, parent=None, color=styles.DIVIDER_COLOR):
        super(QHLine, self).__init__(parent)
        self.setFrameShape(QtWidgets.QFrame.HLine)
        self.setFrameShadow(QtWidgets.QFrame.Plain)
        self.setLineWidth(0)
        self.setMidLineWidth(3)
        self.setContentsMargins(0, 0, 0, 0)

        color = QtGui.QColor(color)

        self.setColor(color)

    def setColor(self, color):
        pal = self.palette()
        pal.setColor(QtGui.QPalette.WindowText, color)
        self.setPalette(pal)


import operator
import types
import importlib

from vent.alarm import AlarmType, AlarmSeverity
from vent.common.message import SensorValues
from vent.common.values import ValueName

def get_alarm_manager():
    try:
        return Alarm_Manager()
    except:
        manager_module = importlib.import_module('vent.alarm.alarm_manager')
        globals()['Alarm_Manager'] = getattr(manager_module, 'Alarm_Manager')
        return Alarm_Manager()


class Condition(object):
    """
    Base class for specifying alarm test conditions

    Need to be able to condition alarms based on
    * value ranges
    * value ranges & durations
    * levels of other alarms


    Attributes:
        manager (:class:`vent.alarm.alarm_manager.Alarm_Manager`): alarm manager, used to get status of alarms
        _child (:class:`Condition`): if another condition is added to this one, store a reference to it
        """

    def __init__(self, depends: dict = None, *args, **kwargs):
        """

        Args:
            depends (list, dict): a list of, or a single dict::

                {'value_name':ValueName,
                'value_attr': attr in ControlMessage,
                 'condition_attr',
                 optional: transformation: callable)
                that declare what values are needed to update
            *args:
            **kwargs:
        """

        self._manager = None
        self._child = None
        self._check = None
        self.depends = depends

    @property
    def manager(self):
        if self._manager is None:
            self._manager = get_alarm_manager()
        return self._manager

    def check(self, sensor_values):
        raise NotImplementedError("Every condition needs to override check!!")

    def reset(self):
        """
        If a condition is stateful, need to provide some method of resetting the state
        """
        raise NotImplementedError("every condition needs to override reset!")

    def __add__(self, other):
        """
        Add another :class:`Condition` object to check in series.

        Conditions are evaluated left-to-right, and return if any along the sequence is False

        Args:
            other (:class:`Condition`)
        """
        # can't just add any ole apples n oranges
        assert(issubclass(type(other), Condition))

        if self._child is None:
            # if something hasn't been added to us yet...
            # claim our child
            self._child = other

            # override our check method so we check recursively
            # make a quick backup first tho yno
            self._check = self.check

            def new_check(self, sensor_values):
                if not self._check(sensor_values):
                    # if our stashed condition check is false,
                    # return immediately
                    return False
                else:
                    # otherwise call check (potentially recursively)
                    return self._child.check(sensor_values)

            # use python types to programmatically reassign method
            self.check = types.MethodType(new_check, self)

        else:
            # if we have already had something added to us,
            # add it to our child instead, (also potentially recursively)
            self._child = self._child + other

        return self


class ValueCondition(Condition):
    """
    value is greater or lesser than some max/min
    """

    def __init__(self,
                 value_name: ValueName,
                 limit: (int, float),
                 mode: str,
                 *args, **kwargs):
        """

        Args:
            value_name (ValueName): Which value to check
            limit (int, float): value to check against
            mode ('min', 'max'): whether the limit is a minimum or maximum
            *args:
            **kwargs:
        """
        super(ValueCondition, self).__init__(*args, **kwargs)

        # self.arguments = [value_name]
        self.value_name = value_name
        self.limit = limit

        self._mode = None
        self.operator = None
        self.mode = mode

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode):
        assert(mode in ('min', 'max'))
        if mode == 'min':
            # if we're a minimum, True (raise alarm) if value is less than limit
            self.operator = operator.lt
        elif mode == 'max':
            self.operator = operator.gt
        else:
            raise ValueError('needs to be max or min')
        self._mode = mode

    def check(self, sensor_values):
        assert(isinstance(sensor_values, SensorValues))
        return self.operator(sensor_values[self.value_name], self.limit)

    def reset(self):
        """
        not stateful, do nothing.
        """
        pass


class CycleValueCondition(ValueCondition):
    """
    value goes out of range for a specific number of breath cycles

    Attributes:
        _start_cycle (int): The breath cycle where the
        _mid_check (bool): whether a value has left the acceptable range and we are counting consecutive breath cycles
    """

    def __init__(self, n_cycles, *args, **kwargs):
        super(CycleValueCondition, self).__init__(*args, **kwargs)
        self._n_cycles = None
        self.n_cycles = n_cycles

        self._start_cycle = 0
        self._mid_check = False

    @property
    def n_cycles(self):
        return self._n_cycles

    @n_cycles.setter
    def n_cycles(self, n_cycles):
        if not isinstance(n_cycles, int):
            n_cycles = int(round(n_cycles))
        assert(n_cycles>0)
        self._n_cycles = n_cycles

    def check(self, sensor_values):
        # first check if we are outside of the range
        if super(CycleValueCondition, self).check(sensor_values):

            breath_cycle = sensor_values.breath_count
            # if we're currently in a consecutive set of out-of-range alarms..
            # note: doing it this way because we *dont* want to alarm if there are
            # in-range values seen in the waiting period, but we *do* want to
            # alarm if we miss a value from a breath cycle but haven't seen any
            # in-range values.
            if self._mid_check:
                # if we have progressed the required number of cycles...
                if breath_cycle >= self._start_cycle + self.n_cycles:
                    return True
                else:
                    return False
            else:
                # otherwise, this is the first time we've gone out of bounds
                self._mid_check = True
                self._start_cycle = breath_cycle
                # don't check yet, n_cycles must > 0
                return False

        else:
            # if we're not outside the range, false.
            # reset the flag that says we're inside a check
            self._mid_check = False
            return False

    def reset(self):
        self._mid_check = False
        self._start_cycle = 0


class TimeValueCondition(ValueCondition):
    """
    value goes out of range for specific amount of time
    """

    def __init__(self, time, *args, **kwargs):
        """

        Args:
            time (float): number of seconds value must be out of range
            *args:
            **kwargs:
        """
        super(TimeValueCondition, self).__init__(*args, **kwargs)


        self.time = time

    def check(self, sensor_values):
        pass

    def reset(self):
        pass

class AlarmSeverityCondition(Condition):

    def __init__(self,
                 alarm_type: AlarmType,
                 severity: AlarmSeverity,
                 mode: str = 'min',
                 *args, **kwargs):
        """

        Args:
            alarm_type:
            severity:
            mode (str): one of 'min', 'equals',  or 'max'.
                'min' returns true if the alarm is at least this value
                (note the difference from ValueCondition which returns true if the alarm is less than..)
                and vice versa for 'max'.

                .. note::

                    'min' and 'max' use >= and <= rather than > and <

            *args:
            **kwargs:
        """
        super(AlarmSeverityCondition, self).__init__(*args, **kwargs)


        self.alarm_type = alarm_type
        self.severity = severity

        self._mode = None
        self.operator = None
        self.mode = mode

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode):
        assert(mode in ('min', 'eq', 'max'))
        if mode == 'min':
            # if we're a minimum, True (raise alarm) if value is less than limit
            self.operator = operator.ge
        elif mode == 'eq':
            self.operator = operator.eq
        elif mode == 'max':
            self.operator = operator.le
        else:
            raise ValueError(f'needs to be max or min, got {mode}')
        self._mode = mode

    def check(self, sensor_values):
        alarm_severity = self.manager.get_alarm_severity(self.alarm_type)
        return self.operator(alarm_severity.value, self.severity.value)

    def reset(self):
        pass


class CycleAlarmSeverityCondition(AlarmSeverityCondition):
    """
    alarm goes out of range for a specific number of breath cycles

    .. todo::

        note that this is exactly the same as CycleValueCondition. Need to do the multiple inheritance thing

    Attributes:
        _start_cycle (int): The breath cycle where the
        _mid_check (bool): whether a value has left the acceptable range and we are counting consecutive breath cycles
    """

    def __init__(self, n_cycles, *args, **kwargs):
        super(CycleAlarmSeverityCondition, self).__init__(*args, **kwargs)
        self._n_cycles = None
        self.n_cycles = n_cycles

        self._start_cycle = 0
        self._mid_check = False

    @property
    def n_cycles(self):
        return self._n_cycles

    @n_cycles.setter
    def n_cycles(self, n_cycles):
        if not isinstance(n_cycles, int):
            n_cycles = int(round(n_cycles))
        assert(n_cycles>0)
        self._n_cycles = n_cycles

    def check(self, sensor_values):
        # first check if we are outside of the range
        if super(CycleAlarmSeverityCondition, self).check(sensor_values):

            breath_cycle = sensor_values.breath_count
            # if we're currently in a consecutive set of out-of-range alarms..
            # note: doing it this way because we *dont* want to alarm if there are
            # in-range values seen in the waiting period, but we *do* want to
            # alarm if we miss a value from a breath cycle but haven't seen any
            # in-range values.
            if self._mid_check:
                # if we have progressed the required number of cycles...
                if breath_cycle >= self._start_cycle + self.n_cycles:
                    return True
                else:
                    return False
            else:
                # otherwise, this is the first time we've gone out of bounds
                self._mid_check = True
                self._start_cycle = breath_cycle
                # don't check yet, n_cycles must > 0
                return False

        else:
            # if we're not outside the range, false.
            # reset the flag that says we're inside a check
            self._mid_check = False
            return False

    def reset(self):
        self._mid_check = False
        self._start_cycle = 0